/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/
#define FILECODE "F(302)"


/*
 * task.c --
 *
 *      Task initialization and switching routines between the host
 *      and the monitor. 
 *     
 *      A task switch: 
 *          -saves the EFLAGS,CR0,CR2,CR4, and IDT
 *          -jumps to code on the shared page
 *              which saves the registers, GDT and CR3
 *              which then restores the registers, GDT and CR3
 *          -restores the IDT,CR0,CR2,CR4 and EFLAGS
 *                
 *      This file is pretty much independent of the host OS.
 *      
 */

#ifdef linux
/* Must come before any kernel header file --hpreg */
#   include "driver-config.h"

#   ifdef USE_PERFCTRS_HOSTED
#      include "perfctr.h"
#   endif

#   define EXPORT_SYMTAB
#endif

#include "vmware.h"
#include "modulecall.h"
#include "vmx86.h"
#include "task.h"
#include "vm_asm.h"
#include "cpuid.h"
#include "hostif.h"
/* On Linux, must come before any inclusion of asm/page.h --hpreg */
#include "hostKernel.h"
#if defined(MODULE_X86_64) && defined(WIN32)
#include "vmmon-asm-x86-64.h"
#endif

/*
 *------------------------------------------------------------------------------
 *
 * Task_InitCrosspage
 *
 *    Initialize the crosspage used to switch to the monitor task. 
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

static Bool hostUsesSysenter = FALSE;
int 
Task_InitCrosspage(VMDriver *vm,          // IN
                   InitBlock *initParams) // IN: Initial params from the VM 
{
   Vcpuid vcpuid;

   ASSERT(sizeof (VMCrossPage) < PAGE_SIZE);
   ASSERT(MODULECALL_CROSS_PAGE_LEN==1);
   for (vcpuid = 0; vcpuid < initParams->numVCPUs;  vcpuid++) {
#ifdef MODULE_X86_64
      uint64 cr3reg;
#else
      uint32 cr3reg;
#endif
      LA linearAddr;

      VMCrossPage *p = HostIF_MapCrossPage(vm, (void *)(uintptr_t)
                                           initParams->crosspage[vcpuid], 
                                           &linearAddr);
#ifdef MODULE_X86_64
      MPN crossPageMPN = HostIF_LookupUserMPN(vm, (void *)(uintptr_t)initParams->crosspage[vcpuid]);
      
      if ((int64)crossPageMPN <= 0) {
         return 1;
      }
#endif
      if (p == NULL) {
         return 1;
      }

      vm->crosspage[vcpuid] = p;
      GET_CR3(cr3reg);
#ifdef MODULE_X86_64
      p->hostContext64.context.cr3 = cr3reg;
      p->hostVA64 = (uintptr_t) p;
      /*
       * In order to make the transition to legacy mode, we must first turn
       * paging off on the processor. For this we need to be executing
       * code on an identity mapped page (the crosspage).
       */
      p->crosspageLinearAddr = (LA)MPN_2_MA(crossPageMPN);

      /* Make sure there is no truncation */
      ASSERT((VA)p->crosspageLinearAddr == MPN_2_MA(crossPageMPN));      
      p->hostContext64VA = (uintptr_t)&p->hostContext64;
#else
      p->hostContext.task.cr3 = cr3reg;
      p->crosspageLinearAddr = linearAddr;
      p->hostContextVA = (uintptr_t)&p->hostContext;
#endif      

      {
#ifdef MODULE_X86_64
         /* XXX hack the initial monitor idt */
         p->monContext.idtr.dtr.offset = 0;
         p->monContext.idtr.dtr.limit = sizeof(Gate)*0x100 -1;
#else         
         DTR idtrReg;
         GET_IDT(idtrReg);
         p->monContext.idtr.dtr = idtrReg; /* irrelevant */
#endif
      }
#ifdef MODULE_X86_64
      {
         p->hostContext64.switchGdtr64.offset = p->crosspageLinearAddr + 
            offsetof(VMCrossPage, switchGdt);
         p->hostContext64.switchGdtr64.limit = sizeof(Descriptor) * 4 - 1;
         
         // null descriptor
         p->switchGdt[0].word[0] = 0;
         p->switchGdt[0].word[1] = 0;

         // 64 bit CS descriptor
         Desc_SetDescriptor(&p->switchGdt[1].desc, 0, (VA32)0xfffff, CODE_DESC, 
                            1, 0, 1, 0, 1);
         // Set the L bit to indicate running 64 bit mode.
         p->switchGdt[1].desc.longmode = 1;

         // 32 bit compatibility mode CS descriptor
         Desc_SetDescriptor(&p->switchGdt[2].desc, 0, (VA32)0xfffff, CODE_DESC, 
                            1, 0, 1, 1, 1);
         
         // 32 bit compatibility mode DS descriptor
         Desc_SetDescriptor(&p->switchGdt[3].desc, 0, (VA32)0xfffff, DATA_DESC, 
                            1, 0, 1, 1, 1);

         // Set up sel:offset for jumping to 64 bit code.
         // The code offset will be set up by the emit code in monitorHosted().
         p->hostContext64.jump64Code.selector = MAKE_SELECTOR(1, SELECTOR_GDT, 0);

         // Set up sel:offset for jumping to 32 bit code.
         // The code offset will be set up by the emit code in monitorHosted().
         p->hostContext64.jump32Code.selector = MAKE_SELECTOR(2, SELECTOR_GDT, 0);

         // Set up segment DS for running code in 32bit compability mode.
         p->hostContext64.switchDS = MAKE_SELECTOR(3, SELECTOR_GDT, 0);
      }
#endif
      
      {
         uint16 trReg;
         GET_TR(trReg);
#ifdef MODULE_X86_64
         p->hostContext64.tr = trReg;
#else
         p->hostContext.tr = trReg;
#endif
      }

      p->irqRelocateOffset[0] = IRQ_HOST_INTR1_BASE; 
      p->irqRelocateOffset[1] = IRQ_HOST_INTR2_BASE;
      p->userCallRequest = MODULECALL_USERCALL_NONE;
      p->moduleCallInterrupted = FALSE;
      p->tscAdjustment = 0;
   }
   return 0;
}


static Bool
DisableNMIDelivery(volatile uint32 *regPtr) // IN/OUT
{
   uint32 reg;
   
   reg = *regPtr;
   if ((APIC_LVT_DELVMODE(reg) == APIC_LVT_DELVMODE_NMI) &&
       (! APIC_LVT_ISMASKED(reg))) {
      *regPtr = reg | APIC_LVT_MASK;
      return TRUE;
   }
   return FALSE;
}


static void
DisableNMI(VMDriver *vm,     // IN
           Bool *lint0NMI,   // OUT
           Bool *lint1NMI,   // OUT
           Bool *pcNMI,      // OUT
           Bool *thermalNMI) // OUT 
{
   if (vm->hostAPIC) {
      *lint0NMI = DisableNMIDelivery(&APIC_LINT0_REG(vm->hostAPIC));
      *lint1NMI = DisableNMIDelivery(&APIC_LINT1_REG(vm->hostAPIC));
      *pcNMI = DisableNMIDelivery(&APIC_PC_REG(vm->hostAPIC));

      /*
       * The LVT thermal monitor register was introduced
       * in Pentium 4 and Xeon processors.
       */

      if (APIC_MAX_LVT(vm->hostAPIC) >= 5) {
         *thermalNMI = DisableNMIDelivery(&APIC_THERM_REG(vm->hostAPIC));
      } else {
         *thermalNMI = FALSE;
      }
   }
}


static void
RestoreNMI(VMDriver *vm,    // IN
           Bool lint0NMI,   // IN
           Bool lint1NMI,   // IN
           Bool pcNMI,      // IN
           Bool thermalNMI) // IN
{
#define RestoreNMIDelivery(cond, apicr)				\
   do {								\
      if (cond) {						\
         uint32 reg;						\
	 							\
         reg = apicr;						\
         apicr = reg & ~APIC_LVT_MASK;				\
      }								\
   } while (0)

   RestoreNMIDelivery(lint0NMI, APIC_LINT0_REG(vm->hostAPIC));
   RestoreNMIDelivery(lint1NMI, APIC_LINT1_REG(vm->hostAPIC));
   RestoreNMIDelivery(pcNMI, APIC_PC_REG(vm->hostAPIC));
   RestoreNMIDelivery(thermalNMI, APIC_THERM_REG(vm->hostAPIC));
#undef RestoreNMIDelivery
}


typedef struct SysenterState {
   Selector     cs;
   Bool         isValidEipEsp;
   uint32       esp;
   uint32       eip;
} SysenterState;


/*
 * Sysenter state handling. Writing MSRs is a costly operation, so we
 * try to conserve on unnecessary writes.
 *
 * We assume that if the host has a 0 in its sysenter CS msr, it is
 * completely sysenter unaware. Therefore, we can let the monitor trample
 * on other MSRs as it pleases.
 *
 * The monitor does not know whether the host uses the MSRs. Thus, in
 * order to avoid unnecessary writes to MSR_SYSENTER_CS, we clear
 * this MSR here if the monitor does not use it but the host
 * has set it to a non-zero value.
 */
static INLINE_SINGLE_CALLER void
SaveSysenterState(SysenterState *st,      // IN
                  VMCrossPage *crosspage) // IN
{
   if (hostUsesSysenter) {
      st->cs = (Selector)HostIF_RDMSR(MSR_SYSENTER_CS);
      /* I am not sure this assert is correct. The host could
       * potentially clear MSR_SYSENTER_CS after we compute
       * hostUsesSysenter. --matthias
       */
      ASSERT(st->cs != 0);
      if (crosspage->monUsesSysenter) {
	 /* Save the esp and eip MSRs. Leave MSR_SYSENTER_CS
          * unchanged; the monitor will write the MSRs
          * appropriately.
          */
	 st->isValidEipEsp = TRUE;
	 st->esp = (uint32)HostIF_RDMSR(MSR_SYSENTER_ESP);
	 st->eip = (uint32)HostIF_RDMSR(MSR_SYSENTER_EIP);
      } else {
	 /* Clear MSR_SYSENTER_CS since the monitor is not
          * using sysenter/sysexit.
	  */
	 st->isValidEipEsp = FALSE;
	 HostIF_WRMSR(MSR_SYSENTER_CS, 0);
      }
   }
}


static INLINE_SINGLE_CALLER void
RestoreSysenterState(SysenterState *st,      // IN
                     VMCrossPage* crosspage) // IN
{
   if (hostUsesSysenter) {
      ASSERT(st->cs != 0);
      HostIF_WRMSR(MSR_SYSENTER_CS, st->cs);
      if (st->isValidEipEsp) {
	 HostIF_WRMSR(MSR_SYSENTER_ESP, st->esp);
	 HostIF_WRMSR(MSR_SYSENTER_EIP, st->eip);
      }
   } else {
      if (crosspage->monUsesSysenter) {
	 /*
	  * The monitor smashed multiple registers, but since the host
	  * appears sysenter unaware, we'll do the minimal amount of work
	  * here to ensure that sysenter raises a #GP.
	  */
	 HostIF_WRMSR(MSR_SYSENTER_CS, 0);
      }
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * SwitchToMonitor --
 *
 *      Wrapper that calls code to switch from the host to the monitor.
 *
 *      The wrapper ends with a long jump after having
 *      pushed EAX and EBX on the stack. Compensate for these
 *      two registers, and then for the retval and argument on
 *      on the stack
 *
 *      Note the dummy 4th argument on the stack compensates
 *      for the absence of segment information pushed on a
 *      near jump.
 *
 * Side effects:
 *      None for the module
 *
 * Bugs: EAX is caller-saved, no need to save/restore
 *       should use caller-saved EDX rather than EBX as temp
 *       in Task_Switch()
 *
 *-----------------------------------------------------------------------------
 */
#ifdef MODULE_X86_64

static INLINE_SINGLE_CALLER void
SwitchToMonitor(VMCrossPage *crosspage)
{
   PA monContextPA   = crosspage->crosspageLinearAddr + 
                       offsetof(VMCrossPage, monContext);
   VA32 monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START)
                       + offsetof(VMCrossPage,monContext));
#ifdef linux
   __asm__ __volatile__(
      "pushq %4"       "\n\t"
      "call *%5"       "\n\t"
      :
      : "a" (&crosspage->hostContext64),
        "b" (monContextPA),
        "c" (monContextVA),
        "d" (crosspage->hostSwitchCR3),
        "g" ((uint64)crosspage->hostContext64.context.cs),
        "r" (crosspage->contextSwitchCode)
      : "memory", "cc"
   );
#else
   SWITCH_TO_MONITOR((uint64)&crosspage->hostContext64, monContextPA,
		   monContextVA, crosspage->hostSwitchCR3,
                   (uint64)crosspage->hostContext64.context.cs,
		   (uint64)crosspage->contextSwitchCode);
#endif
}

#else

static INLINE_SINGLE_CALLER void
SwitchToMonitor(VMCrossPage *crosspage)
{
   char *codePtr    = &crosspage->contextSwitchCode[0];
   ContextInfo *src = &(crosspage->hostContext);
   ContextInfo *dst = &(crosspage->monContext);
   /*
    * newDstVA points to the same location as
    * vm->crosspage->monContext, only refers to it
    * in the monitor's virtual address space
    */
   uint32 newDstVA  = (uint32)
      (VPN_2_VA(MODULECALL_CROSS_PAGE_START)+ offsetof(VMCrossPage,monContext));    
#ifdef __GNUC__
   uint32 dummy1;
   uint32 dummy2;

   /* Checked against the Intel manual and GCC --hpreg */
   __asm__ __volatile__(
      "pushl %2"       "\n\t"
      "pushl %3"       "\n\t"
      "pushl %4"       "\n\t"
      "pushl %5"       "\n\t"
      "call *%6"       "\n\t"
      "popl %0"        "\n\t"
      "popl %1"        "\n\t"
      "addl %7, %%esp"
      /* 
       * Unfortunately there is no way to tell to gcc that we are 
       * using 'pushl' and so %esp-relative addressing does not
       * work. To prevent miscompilation we must limit ourselves
       * to registers and immediate values only.
       *
       * We do not have to use early-clobber, as %0 & %1 are used
       * only before %7, and %7 is immediate value.
       */
      : "=b" (dummy1),
        "=a" (dummy2)
      : "g" (newDstVA),
        "ri" (dst),
        "ri" (src),
        "ri" (0),
        "r" (codePtr),
        "i" ((1 + 1 + 3) * 4) // retval + segment + 3 arguments pushed
      : "memory", "cc"
   );
#elif _MSC_VER
   __asm push newDstVA
   __asm push dst
   __asm push src
   __asm push 0
   __asm call codePtr
   __asm pop ebx
   __asm pop eax
   __asm add esp, 20
#else
#   error No compiler defined for SwitchToMonitor
#endif
}

#endif // MODULE_X86_64

/*
 *-----------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state.
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned.
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt.
 *
 *-----------------------------------------------------------------------------
 */

void 
Task_Switch(VMDriver *vm,  // IN
            Vcpuid vcpuid) // IN
{
#ifdef MODULE_X86_64
   uint64      flags, cr0reg, cr2reg, cr4reg, new_cr4;
   uint64      gs64, fs64;
   uint64      drReg;
#else
   DTR         gdtreg;
   uint32      flags,cr0reg, cr2reg, cr4reg, new_cr4;
   uint32      drReg;
#endif
   Descriptor *desc;
   SysenterState sysenterState;
   VA          vAddr;
   Selector    cs, gs, fs;
   uint16 trReg;   
   uint16 ldtreg;
   Bool lint0NMI = FALSE;
   Bool lint1NMI = FALSE;
   Bool pcNMI = FALSE;
   Bool thermalNMI = FALSE;
   VMCrossPage *crosspage = vm->crosspage[vcpuid];
   
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromHost();
#endif
   DisableNMI(vm, &lint0NMI, &lint1NMI, &pcNMI, &thermalNMI);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();

   vm->currentHostCpu[vcpuid] = HostIF_GetCurrentPCPU();
   
   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);
   
   /* Ensure global pages are flushed */
   new_cr4 = cr4reg & ~CR4_PGE;
   SET_CR4(new_cr4);

   crosspage->switchHostCR4 = new_cr4;
   
   /*
    * Save DR7 since we need to disable debug breakpoints during the
    * world switch code.  Save DR6 in order to accomodate the ICEBP
    * instruction.  All other debug registers are saved lazily by the
    * monitor prior to first use.  NOTE: Since monitor is in legacy
    * mode save all DRs for x86_64 here.
    */

   crosspage->hostDRSaved = 0;

#define SAVE_DR_SHADOW(n) \
                GET_DR##n(drReg); \
                crosspage->hostDR[n] = drReg; \
                crosspage->shadowDR[n] = (uint32) drReg; \
                crosspage->hostDRSaved |= (1 << n);

#ifdef MODULE_X86_64
   SAVE_DR_SHADOW(0);
   SAVE_DR_SHADOW(1);
   SAVE_DR_SHADOW(2);
   SAVE_DR_SHADOW(3);
#endif // MODULE_X86_64

   SAVE_DR_SHADOW(6);
   SAVE_DR_SHADOW(7);
   
#undef SAVE_DR_SHADOW
   
   /*
    * Disable debug exceptions during the switch.  Note: If
    * the DR7 GD bit is on, a trap occurred at the GET_DR
    * above, and this process is most likely dead.
    */

   if (UNLIKELY(crosspage->hostDR[7] & DR7_ENABLED)) {
      SET_DR7(DR7_DEFAULT);
      crosspage->shadowDR[7] = DR7_DEFAULT;
   }

#ifdef MODULE_X86_64
   GET_GDT64(crosspage->hostContext64.gdtr64.dtr);
#else
   GET_GDT(gdtreg);
#endif
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
#ifdef MODULE_X86_64
   gs64 = GET_GS64();
   fs64 = GET_FS64();
#endif
   gs = GET_GS();
   fs = GET_FS();
   
   SaveSysenterState(&sysenterState, crosspage);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    */
   
   GET_TR(trReg);
#ifdef MODULE_X86_64
   crosspage->hostContext64.tr = trReg;
#else
   crosspage->hostContext.tr = trReg;
#endif
   if (trReg) {
#ifdef MODULE_X86_64
      vAddr = HOST_KERNEL_LA_2_VA(crosspage->hostContext64.gdtr64.dtr.offset) + trReg;
#else      
      vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
#endif     
     desc = (Descriptor *)(vAddr);
     if (Desc_Type(desc) == TASK_DESC_BUSY) {
       Desc_SetType(desc, TASK_DESC);
     }
   }


   /*
    * Save the CS on this side
    */
   cs = GET_CS();
#ifdef MODULE_X86_64
   crosspage->hostContext64.context.cs = cs;
#else
   *(uint16*) (&crosspage->hostContext.jump[4]) = cs;
   crosspage->hostContext.task.cs = cs;
#endif   
   /*
    * See comment for SwitchToMonitor()
    */
   SwitchToMonitor(crosspage);

   RestoreSysenterState(&sysenterState, crosspage);
   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   
   SET_LDT(ldtreg);

   /* restore fs/gs must come before 64 bit fs/gs restore */
   SET_FS(fs);
   SET_GS(gs);
#ifdef MODULE_X86_64
   SET_FS64(fs64);
   SET_GS64(gs64);
#endif

#ifdef MODULE_X86_64
#define CAST_HOST_DR(x) (x)
#else
#define CAST_HOST_DR(x) ((uint32) x)
#endif

#define RESTORE_DR(n) if ((crosspage->hostDRSaved & (1 << n)) && \
                          (crosspage->hostDR[n] != crosspage->shadowDR[n])) { \
                          drReg = CAST_HOST_DR(crosspage->hostDR[n]); \
                          SET_DR##n(drReg); \
                      }

   RESTORE_DR(0);
   RESTORE_DR(1);
   RESTORE_DR(2);
   RESTORE_DR(3);
   RESTORE_DR(6);
   RESTORE_DR(7);

#undef RESTORE_DR
#undef CAST_HOST_DR
   
   ASSERT_NO_INTERRUPTS();

   if (crosspage->moduleCallType == MODULECALL_INTR) {
      /*
       * Note we must do the RAISE_INTERRUPT before ever enabling
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw).
       * Note2: RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement.
       */

#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)

      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.
#ifdef MODULE_X86_64
	 IRQ_INT32(0x00);
#else
	 /*
	  * Pass Machine Check Exception (Interrupt 0x12) to the host.
	  * See bug #45286 for details.
	  */
	 IRQ_INT(0x12);
#endif
	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }
   }
   
   vm->currentHostCpu[vcpuid] = INVALID_HOST_CPU;

   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint0NMI, lint1NMI, pcNMI, thermalNMI);
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromMonitor();
#endif
}


/*
 *----------------------------------------------------------------------
 * Task_ProbeSysenterMSR --
 *
 *    This function computes sysenterMSR. Note that we assume that if a
 *    host is ever going to use the sysenter MSRs, they will have been
 *    initialized here; if some future host lazily sets up the sysenter
 *    MSRs, this code will cause trouble.
 *
 *----------------------------------------------------------------------
 */
void
Task_ProbeSysenterMSR(void)
{
   CPUOem cpuOem = CPUID_GetOEM();
   uint32 features = CPUID_GetFeatures();
   uint32 version = CPUID_GetVersion();

   /*
    * Some old intel CPUs set the SEP bit, even though they are not
    * capable of sysenter. In particular, CPUs with family of 6, but
    * with the whole version string < 0x633 (family, model, stepping).
    * This check is also performed in vmx/vmcore/cpuid.c.
    */
   if (cpuOem == CPUID_OEM_INTEL && CPUID_FAMILY(version) == 6 &&
       (CPUID_MODEL(version) < 3 ||
        (CPUID_MODEL(version) == 3 && CPUID_STEPPING(version) < 3))) {

      Warning("VMMON: CPU too old to use sysenter (ok).\n");
   } else if (features & CPUID_FEATURE_COMMON_SEP) {
      uint64 tmpMSR;
      /* The CPU appears to support sysenter. Does the host actually
       * use it?
       */
      tmpMSR = HostIF_RDMSR(MSR_SYSENTER_CS);
      hostUsesSysenter = (tmpMSR != 0);
   }
}
