/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved.
 * -- VMware Confidential
 * **********************************************************/

/*
 * vm_asm_x86_64.h
 *
 *	x86-64 asm macros
 */

#ifndef _VM_ASM_X86_64_H_
#define _VM_ASM_X86_64_H_

#define INCLUDE_ALLOW_MODULE
#include "includeCheck.h"

#include "x86.h"

/*
 * This file contains inline assembly routines used by x86_64 code.
 * Currently only vmmon and vmnet will use routines defined in this
 * file.
 * 
 */

#ifdef __GNUC__
/* Checked against the Intel manual and GCC --hpreg */
#define _BUILD_SET_DTR(table)  \
static INLINE void             \
_Set_ ## table ## dt(DTR *dtr) \
{                              \
   __asm__(                    \
      "l" #table "dt %0"       \
      :                        \
      : "m" (*dtr)             \
   );                          \
}

#define _BUILD_SET_DTR64(table)    \
static INLINE void                 \
_Set_ ## table ## dt64(DTR64 *dtr) \
{                                  \
   __asm__(                        \
      "l" #table "dt %0"           \
      :                            \
      : "m" (*dtr)                 \
   );                              \
}

_BUILD_SET_DTR(i)
#define SET_IDT(var) do { \
   _Set_idt(&(var));      \
} while (0)
_BUILD_SET_DTR(g)
#define SET_GDT(var) do { \
   _Set_gdt(&(var));      \
} while (0)
_BUILD_SET_DTR64(i)
#define SET_IDT64(var) do { \
   _Set_idt64(&(var));      \
} while (0)
_BUILD_SET_DTR64(g)
#define SET_GDT64(var) do { \
   _Set_gdt64(&(var));      \
} while (0)


/* Checked against the Intel manual and GCC --hpreg

   volatile because the [IG]DT can change without the compiler knowing it
   (when we use l[ig]dt). */
#define _BUILD_GET_DTR(table)  \
static INLINE void             \
_Get_ ## table ## dt(DTR *dtr) \
{                              \
   __asm__ __volatile__(       \
      "s" #table "dt %0"       \
      : "=m" (*dtr)            \
   );                          \
}

#define _BUILD_GET_DTR64(table)    \
static INLINE void                 \
_Get_ ## table ## dt64(DTR64 *dtr) \
{                                  \
   __asm__ __volatile__(           \
      "s" #table "dt %0"           \
      : "=m" (*dtr)                \
   );                              \
}

_BUILD_GET_DTR(i);
#define GET_IDT(var) do { \
   _Get_idt(&(var));      \
} while (0)
_BUILD_GET_DTR(g);
#define GET_GDT(var) do { \
   _Get_gdt(&(var));      \
} while (0)
_BUILD_GET_DTR64(i);
#define GET_IDT64(var) do { \
   _Get_idt64(&(var));      \
} while (0)
_BUILD_GET_DTR64(g);
#define GET_GDT64(var) do { \
   _Get_gdt64(&(var));      \
} while (0)


/* Checked against the Intel manual and GCC --hpreg */
static INLINE void
SET_LDT(Selector s)
{
   __asm__(
      "lldt %0"
      :
      : "rm" (s)
   );
}


/* Checked against the Intel manual and GCC --hpreg

   volatile because the LDT can change without the compiler knowing it
   (when we use lldt). */
static INLINE Selector
_Get_ldt(void)
{
   Selector result;

   __asm__ __volatile__(
      "sldt %0"
      : "=rm" (result)
   );

   return result;
}

#define GET_LDT(var) do { \
   var = _Get_ldt();      \
} while (0)

/* Not checked against the Intel manual and GCC yet --slava */
#define _BUILD_SET_R(func, reg) \
static INLINE void              \
func(uint64 r)                  \
{                               \
   __asm__(                     \
      "movq %0, %%" #reg        \
      :                         \
      : "r" (r)                 \
      : "memory"                \
   );                           \
}

_BUILD_SET_R(SET_CR0, cr0)
_BUILD_SET_R(SET_CR2, cr2)
_BUILD_SET_R(SET_CR3, cr3)
_BUILD_SET_R(SET_CR4, cr4)

_BUILD_SET_R(SET_DR0, dr0)
_BUILD_SET_R(SET_DR1, dr1)
_BUILD_SET_R(SET_DR2, dr2)
_BUILD_SET_R(SET_DR3, dr3)
_BUILD_SET_R(SET_DR6, dr6)
_BUILD_SET_R(SET_DR7, dr7)


/* Not checked against the Intel manual and GCC yet --slava

   volatile because CRs and DRs can change without the compiler knowing it
   (when there is a page fault, when a breakpoint occurs, and moreover it seems
   there is no way to teach gcc that smsw clobbers cr0 for example). */
#define _BUILD_GET_R(reg)   \
static INLINE uint64        \
_Get_ ## reg(void)          \
{                           \
   uint64 result;           \
                            \
   __asm__ __volatile__(    \
      "movq %%" #reg ", %0" \
      : "=r" (result)       \
   );                       \
                            \
   return result;           \
}

_BUILD_GET_R(cr0)
#define GET_CR0(var) do { \
   var = _Get_cr0();      \
} while (0)
_BUILD_GET_R(cr2)
#define GET_CR2(var) do { \
   var = _Get_cr2();      \
} while (0)
_BUILD_GET_R(cr3)
#define GET_CR3(var) do { \
   var = _Get_cr3();      \
} while (0)
_BUILD_GET_R(cr4)
#define GET_CR4(var) do { \
   var = _Get_cr4();      \
} while (0)

_BUILD_GET_R(dr0)
#define GET_DR0(var) do { \
   var = _Get_dr0();      \
} while (0)
_BUILD_GET_R(dr1)
#define GET_DR1(var) do { \
   var = _Get_dr1();      \
} while (0)
_BUILD_GET_R(dr2)
#define GET_DR2(var) do { \
   var = _Get_dr2();      \
} while (0)
_BUILD_GET_R(dr3)
#define GET_DR3(var) do { \
   var = _Get_dr3();      \
} while (0)
_BUILD_GET_R(dr6)
#define GET_DR6(var) do { \
   var = _Get_dr6();      \
} while (0)
_BUILD_GET_R(dr7)
#define GET_DR7(var) do { \
   var = _Get_dr7();      \
} while (0)


/* Checked against the Intel manual and GCC --hpreg */
#define _BUILD_SET_SEG(func, reg) \
static INLINE void                \
func(Selector s)                  \
{                                 \
   __asm__ __volatile__(          \
      "movw %0, %%" #reg          \
      :                           \
      : "rm" (s)                  \
   );                             \
}

_BUILD_SET_SEG(SET_DS, ds)
_BUILD_SET_SEG(SET_ES, es)
_BUILD_SET_SEG(SET_FS, fs)
_BUILD_SET_SEG(SET_GS, gs)
_BUILD_SET_SEG(SET_SS, ss)


/* Checked against the Intel manual and GCC --hpreg

   volatile because the content of CS can change without the compiler knowing
   it (when we use call gates). */
#define _BUILD_GET_SEG(func, reg) \
static INLINE Selector            \
func(void)                        \
{                                 \
   Selector result;               \
                                  \
   __asm__ __volatile__(          \
      "movw %%" #reg ", %0"       \
      : "=rm" (result)            \
   );                             \
                                  \
   return result;                 \
}

_BUILD_GET_SEG(GET_CS, cs)
_BUILD_GET_SEG(GET_DS, ds)
_BUILD_GET_SEG(GET_ES, es)
_BUILD_GET_SEG(GET_FS, fs)
_BUILD_GET_SEG(GET_GS, gs)
_BUILD_GET_SEG(GET_SS, ss)



/* Checked against the Intel manual and GCC --hpreg */
static INLINE void
SET_TR(Selector s)
{
   __asm__(
      "ltr %0"
      :
      : "rm" (s)
   );
}


/* Checked against the Intel manual and GCC --hpreg

   volatile because the content of TR can change without the compiler knowing
   it (when we use task gates). */
static INLINE Selector
_Get_TR(void)
{
   Selector result;

   __asm__ __volatile__(
      "str %0"
      : "=rm" (result)
   );

   return result;
}

#define GET_TR(var) do { \
   var = _Get_TR();      \
} while (0)


/* Checked against the Intel manual and GCC --hpreg */
static INLINE void
_Set_rflags(uint64 f)
{
   __asm__(
      "push %0" "\n\t"
      "popf"
      :
      : "g" (f)
      : "cc"
   );
}

#define RESTORE_FLAGS _Set_rflags


/* Checked against the Intel manual and GCC --hpreg

   volatile because gcc 2.7.2.3 doesn't know when eflags are modified (it
   seems to ignore the "cc" clobber). gcc 2.95.2 is ok: it optimize 2
   successive calls only if there is no instructions in between. */
static INLINE uint64
_Get_rflags(void)
{
   uint64 result;

   __asm__ __volatile__(
      "pushf"  "\n\t"
      "pop %0"
      : "=rm" (result)
   );

   return result;
}

#define SAVE_FLAGS(var) do { \
   var = _Get_rflags();      \
} while (0)

/* Checked against the Intel manual and GCC --hpreg */
static INLINE void
CLTS(void)
{
   __asm__("clts");
}


/* Beginning of the section whose correctness has NOT been checked */
#define FNCLEX()         __asm__("fnclex" ::);

/* TLB_INVALIDATE_xxx checked  yet --slava */
#define TLB_INVALIDATE_PAGE(_addr) do { \
     __asm__ __volatile__("invlpg %0": :"m" (*(char *) (_addr)):"memory"); \
} while (0)

#define TLB_INVALIDATE_PAGE_OFF_FS(_addr) do { \
     __asm__ __volatile__("fs; invlpg %0": :"m" (*(char *) (_addr)):"memory"); \
} while (0)


#if (defined(VMKERNEL) || defined(VMKERNEL_MODULE)) && defined(VMX86_DEBUG)
/* {ENABLE|CLEAR}_INTERRUPTS defined in vmkernel_dist.h */
#include "vmkernel_dist.h"
#else
#define ENABLE_INTERRUPTS() __asm__ __volatile__ ("sti": : :"memory")
#define CLEAR_INTERRUPTS()  __asm__ __volatile__ ("cli": : :"memory")
#endif
#define RAISE_INTERRUPT(_x)  __asm__ __volatile__("int %0" :: "g" (_x))
#define RETURN_FROM_INT()   __asm__ __volatile__("iret" :: ) 

#define NO_INTERRUPTS_BEGIN()	do { \
                                   uint64 _flags; \
                                   SAVE_FLAGS(_flags); \
                                   CLEAR_INTERRUPTS();

#define NO_INTERRUPTS_END()	   RESTORE_FLAGS(_flags); \
                                } while(0)

#elif _MSC_VER  /* !__GNUC__ */

#define GET_CR0(_reg) do { \
   (_reg) = __readcr0();   \
} while(0)
#define SET_CR0(_reg) __writecr0(_reg)
   
#define GET_CR2(_reg) do { \
   (_reg) = __readcr2();   \
} while(0)

#define GET_CR3(_reg) do { \
   (_reg) = __readcr3();   \
} while(0)

#define SET_CR3(_reg) __writecr3(_reg)

#define GET_CR4(_reg) do { \
   (_reg) = __readcr4();   \
} while(0)

#define SET_CR4(_reg) __writecr4(_reg)

#define GET_CR8(_reg) do { \
   (_reg) = __readcr8();   \
} while(0)

#define SET_CR8(_reg) __writecr8(_reg)
#else
#error No compiler defined for get/set
#endif /* !__GNUC__ && !_MSC_VER */


#ifdef __GNUC__
/* Checked against the Intel manual and GCC --hpreg

   volatile because the msr can change without the compiler knowing it
   (when we use wrmsr). */
static INLINE uint64 __GET_MSR(int cx)
{
   uint64 msr;
   uint32 low, high;

   __asm__ __volatile__(
      "rdmsr"
      : "=a" (low), "=d" (high)
      : "c" (cx)
   );

   msr = ((uint64)low) | (((uint64)high) << 32);
   return msr;
}
#elif _MSC_VER
static INLINE uint64 __GET_MSR(int input)
{
   return __readmsr(input);
}   
static INLINE void __SET_MSR(int cx, uint64 val)
{
  __writemsr(cx, val);
}
#else
#error
#endif

#ifdef __GNUC__
static INLINE void __SET_MSR(int cx, uint64 val)
{
   uint32 low, high;
   low = val;
   high = val >> 32;
   __asm__ __volatile__(
      "wrmsr"
      : /* no outputs */			
      : "a" (low), "d"(high),
        "c" (cx)
    );
}
#endif
static INLINE void SET_FS64(uint64 fs64)
{
   __SET_MSR(0xc0000100, fs64);
}

static INLINE void SET_GS64(uint64 gs64)
{
   __SET_MSR(0xc0000101, gs64);
}

static INLINE uint64 GET_FS64(void)
{
   return __GET_MSR(0xc0000100);
}

static INLINE uint64 GET_GS64(void)
{
   return __GET_MSR(0xc0000101);
}

#if defined(__GNUC__)
static INLINE Bool
INTERRUPTS_ENABLED(void)
{
   uintptr_t flags;
   SAVE_FLAGS(flags);
   return ((flags & EFLAGS_IF) != 0);
}
#endif

/*
 *-----------------------------------------------------------------------------
 *
 * ChangeAddrSpaceWithoutDelay --
 *   Flushes the TLB and changes the address space by assigning CR3
 *   to the specified value.  Does not include delay for working around
 *   PIII CR3 setting problem (see BUG 6176). Work around is only necessary
 *   if monitor uses global pages. 
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   Flushes TLB and changes the address space.
 *
 *-----------------------------------------------------------------------------
 */
static INLINE void
ChangeAddrSpaceWithoutDelay(uint32 cr3value)
{
   SET_CR3(cr3value);
}
#endif
