/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved.
 * -- VMware Confidential
 * **********************************************************/

/*
 * vm_assert.h --
 *
 *      The basic assertion facility for all VMware code.
 *
 *      For proper use, see
 *      http://vmweb.vmware.com/~mts/WebSite/guide/programming/asserts.html
 */

#ifndef VM_ASSERT_H
#define VM_ASSERT_H

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_VMM
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_DISTRIBUTE
#include "includeCheck.h"

#include "vm_basic_types.h"

/*
 * Compact filecodes for the monitor.
 */

#ifdef FILECODEINT
#define FILECODE XSTR(F(FILECODEINT))
#endif

#ifndef FILECODEINT
#ifdef MONITOR
#error Monitor must use FILECODEINT
#else
#define FILECODEINT 0
#endif
#endif


#if defined(VMX86_DEVEL)
#define FILECODESTRING  __FILE__
#else
#ifndef FILECODE
#error <FILECODE> not defined.  Update ".../support/scripts/filecodename.prl".
#endif
#define FILECODESTRING FILECODE 
#endif

#define NULL_PANIC_LOCATION ((const char *) -1)

EXTERN NORETURN void Panic(char *, ...) PRINTF_DECL(1, 2);
EXTERN NORETURN void PanicLocation(const char *, char *, ...)
                     PRINTF_DECL(2, 3);

EXTERN void Warning(char *, ...) PRINTF_DECL(1, 2);
EXTERN void Log(char *, ...) PRINTF_DECL(1, 2);
EXTERN void WarningThrottled(uint32 *count, char *, ...) PRINTF_DECL(2, 3);
EXTERN void LogThrottled(uint32 *count, char *, ...) PRINTF_DECL(2, 3);
/* DB family:  messages which are parsed by logfile database system */
#define WarningDB Warning
#define LogDB Log
#define WarningThrottledDB WarningThrottled
#define LogThrottledDB LogThrottled



#define AssertPanicStr             "PANIC %s:%d\n"
#define AssertNotImplStr           "NOT_IMPLEMENTED %s:%d\n"
#define AssertBugStr               "ASSERT %s:%d bugNr=%d\n"
#define AssertNotImplBugStr        "NOT_IMPLEMENTED %s:%d bugNr=%d\n"
#define AssertPanicBugStr          "PANIC %s:%d bugNr=%d\n"
#define AssertNotReachedBugStr     "NOT_REACHED %s:%d bugNr=%d\n"
#define AssertAssertFailedStr      "ASSERT %s:%d\n"
#define AssertNotReachedStr        "NOT_REACHED %s:%d\n"
#define AssertAssertLengthStr      "ASSERT_LENGTH %s:%d r=0x%x e=0x%x\n"
#define AssertUnexpectedWarningStr "UNEXPECTED %d (warning) %s:%d\n"
#define AssertNotTestedWarningStr  "NOT_TESTED (warning) %s:%d\n"
#define AssertMemAllocFailedStr    "MEM_ALLOC %s:%d\n"
#define AssertMemAllocFailedBugStr "MEM_ALLOC %s:%d bugNr=%d\n"

#ifdef MONITOR
extern char Assert_AssertLength[];
extern char Assert_UnexpectedWarning[];
extern char Assert_NotTestedWarning[];
#else /* MONITOR */
#define Assert_Panic AssertPanicStr
#define Assert_NotImpl AssertNotImplStr
#define Assert_AssertFailedBug AssertBugStr
#define Assert_NotImplBug AssertNotImplBugStr
#define Assert_PanicBug AssertPanicBugStr
#define Assert_NotReachedBug AssertNotReachedBugStr
#define Assert_AssertFailed AssertAssertFailedStr
#define Assert_NotReached AssertNotReachedStr
#define Assert_AssertLength AssertAssertLengthStr
#define Assert_UnexpectedWarning AssertUnexpectedWarningStr
#define Assert_NotTestedWarning AssertNotTestedWarningStr
#define Assert_MemAlloc AssertMemAllocFailedStr
#define Assert_MemAllocBug AssertMemAllocFailedBugStr
#endif

/*
 * Compilation options
 */

#if defined(MONITOR)
#define VMX86_DENSE_ASSERTS
EXTERN NORETURN void PanicNotReached(const char *location);
EXTERN NORETURN void PanicNotImplemented(const char *location);
EXTERN NORETURN void PanicAssert(const char *location);
EXTERN NORETURN void PanicPanic(const char *location);
EXTERN NORETURN void PanicNotReachedBug(const char *location, int bug);
EXTERN NORETURN void PanicNotImplementedBug(const char *location, int bug);
EXTERN NORETURN void PanicAssertBug(const char *location, int bug);
EXTERN NORETURN void PanicPanicBug(const char *location, int bug);
#endif


/*
 * Internal macros
 */

#ifdef VMX86_DENSE_ASSERTS // {

   #define WITH_LOCATION(block) do { \
      static char _location_[] __attribute__((section (".locations"))) = \
           __FILE__ ":" XSTR(__LINE__); \
      block \
   } while (FALSE)

   #define _ASSERT_PANIC(bug, cond, str, func) do { \
      if (UNLIKELY(!(cond))) { \
	 WITH_LOCATION( \
	    if ((bug) == 0) { \
	       (func)(_location_); \
	    } else { \
	       (func##Bug)(_location_, bug); \
	    } \
	 ); \
      } \
   } while (FALSE)

#else // } {

   #define _ASSERT_PANIC(bug, cond, str, func) do { \
      if (UNLIKELY(!(cond))) { \
	 if ((bug) == 0) { \
	    Panic(str, FILECODESTRING, __LINE__); \
	 } else { \
	    Panic(str##Bug, FILECODESTRING, __LINE__, bug); \
	 } \
      } \
   } while (FALSE)

#endif // }

#define _ASSERT_WARNING(_c, _s) do {           \
   if (UNLIKELY(! (_c))) {                     \
      Warning((_s), FILECODESTRING, __LINE__); \
   }                                           \
} while (0)

#define _ASSERT_LOG(_c, _s) do {           \
   if (UNLIKELY(! (_c))) {                 \
      Log((_s), FILECODESTRING, __LINE__); \
   }                                       \
} while (0)

#define _ASSERT_WARNING_BUG(_c, _s, _b) do {         \
   if (UNLIKELY(! (_c))) {                           \
      Warning((_s), (_b), FILECODESTRING, __LINE__); \
   }                                                 \
} while (0)

#define _ASSERT_LOG_BUG(_c, _s, _b) do {         \
   if (UNLIKELY(! (_c))) {                       \
      Log((_s), (_b), FILECODESTRING, __LINE__); \
   }                                             \
} while (0)

#define PANIC() \
   _ASSERT_PANIC(0, FALSE, Assert_Panic, PanicPanic)

#define ASSERT_NOT_IMPLEMENTED(_c) \
   _ASSERT_PANIC(0, _c, Assert_NotImpl, PanicNotImplemented)

#define NOT_IMPLEMENTED() \
   _ASSERT_PANIC(0, FALSE, Assert_NotImpl, PanicNotImplemented)

#define ASSERT_BUG(bug, cond) \
   _ASSERT_PANIC(bug, cond, Assert_AssertFailed, PanicAssert)

#define ASSERT_NOT_IMPLEMENTED_BUG(bug, cond) \
   _ASSERT_PANIC(bug, cond, Assert_NotImpl, PanicNotImplemented)

#define NOT_IMPLEMENTED_BUG(bug) \
   _ASSERT_PANIC(bug, FALSE, Assert_NotImpl, PanicNotImplemented)

#define PANIC_BUG(bug) \
   _ASSERT_PANIC(bug, FALSE, Assert_Panic, PanicPanic)

#define NOT_REACHED_BUG(bug) \
   _ASSERT_PANIC(bug, FALSE, Assert_NotReached, PanicNotReached)
   
#define ASSERT_NOT_IMPLEMENTED_PERROR(_cond) {                    \
   if (UNLIKELY(!(_cond))) {                                      \
      Panic("Unimplemented system call return value %s:%d (%s)\n",\
            FILECODESTRING,__LINE__, Msg_ErrString());            \
   }                                                              \
}

#define ASSERT_MEM_ALLOC(cond) \
   _ASSERT_PANIC(0, cond, Assert_MemAlloc, PanicNotImplemented)

#define LOG_ONCE(_s)                \
   do {                             \
      static Bool logged = FALSE;   \
      if (!logged) {                \
	 Log _s;                    \
         logged = TRUE;             \
      }                             \
   } while (0)


/*
 * Following asserts are present only in the debug versions
 */

#define NOT_TESTED_WARNING()  NOT_TESTED()

#ifdef VMX86_DEBUG // {
#undef ASSERT
#define ASSERT(_c) \
   _ASSERT_PANIC(0, _c, Assert_AssertFailed, PanicAssert)

#define NOT_REACHED() \
   _ASSERT_PANIC(0, FALSE, Assert_NotReached, PanicNotReached)

#define ASSERT_BUG_DEBUGONLY(_bug,_c) ASSERT_BUG(_bug,_c)

#ifdef VMX86_DEVEL
#define ASSERT_LENGTH(_real,_expected) {                                      \
   if (UNLIKELY((_real) != (_expected))) {                                    \
      Panic(Assert_AssertLength,                                              \
            FILECODESTRING, __LINE__, _real,_expected);                       \
   }                                                                          \
}
#else  /* !VMX86_DEVEL */
#define ASSERT_LENGTH(_real,_expected) ASSERT((_real) == (_expected))
#endif /* !VMX86_DEVEL */

#else // VMX86_DEBUG } {

#ifndef WINNT_DDK
#undef ASSERT
#define ASSERT(_c)
#endif

/*
 * Expand NOT_REACHED() as appropriate for each situation.
 *
 * Mainly, we want the compiler to infer the same control-flow
 * information as it would from Panic().  Otherwise, different
 * compilation options will lead to different control-flow-derived
 * errors, causing some make targets to fail while others succeed.
 *
 * VC++ has the __assume() built-in function which we don't trust
 * (see bug 43485); gcc has no such construct; we just panic in
 * userlevel code.  The monitor doesn't want to pay the size penalty
 * (measured at 212 bytes for the release vmm for a minimal infinite
 * loop; panic would cost even more) so it does without and lives
 * with the inconsistency.
 */

#ifdef MONITOR
#define NOT_REACHED()
#else
#define NOT_REACHED() \
   _ASSERT_PANIC(0, FALSE, Assert_NotReached, PanicNotReached)
#endif

#define ASSERT_LENGTH(_real,_expected)
#define ASSERT_BUG_DEBUGONLY(_bug,_c)
#endif // !VMX86_DEBUG }


/*
 * LOG_UNEXPECTED(bugNr) 
 */
#ifdef VMX86_DEBUG
#if defined(VMX86_BETA)
#define ASSERT_LOG_UNEXPECTED(_bugNr, _c) \
   _ASSERT_LOG_BUG(_c, Assert_UnexpectedWarning, _bugNr)
#define LOG_UNEXPECTED(_bugNr) \
   _ASSERT_LOG_BUG(0,  Assert_UnexpectedWarning, _bugNr)
#elif defined(VMX86_RELEASE)
#error
#else
#define ASSERT_LOG_UNEXPECTED(_bugNr, _c) \
   _ASSERT_WARNING_BUG(_c, Assert_UnexpectedWarning, _bugNr)
#define LOG_UNEXPECTED(_bugNr) \
   _ASSERT_WARNING_BUG(0,  Assert_UnexpectedWarning, _bugNr)
#endif
#else  /* VMX86_DEBUG */
#define ASSERT_LOG_UNEXPECTED(_bugNr, _c)
#define LOG_UNEXPECTED(_bugNr)
#endif /* VMX86_DEBUG */

/*
 * Different NOT_TESTED() for VMX86_BETA 
 */
#ifdef VMX86_DEBUG
#if defined(VMX86_BETA)
#define ASSERT_NOT_TESTED(_c) _ASSERT_LOG(_c,Assert_NotTestedWarning)
#define NOT_TESTED()          _ASSERT_LOG(0, Assert_NotTestedWarning)
#elif defined(VMX86_RELEASE)
#error
#else
#define ASSERT_NOT_TESTED(_c) _ASSERT_WARNING(_c,Assert_NotTestedWarning)
#define NOT_TESTED()          _ASSERT_WARNING(0, Assert_NotTestedWarning)
#endif

#define NOT_TESTED_ONCE()                                               \
   do {                                                                 \
      static Bool alreadyPrinted = FALSE;                               \
      if (UNLIKELY(!alreadyPrinted)) { alreadyPrinted = TRUE; NOT_TESTED(); } \
   } while (0)

#define NOT_TESTED_1024()                                               \
   do {                                                                 \
      static uint16 count = 0;                                          \
      if (UNLIKELY(count == 0)) { NOT_TESTED(); }                       \
      count = (count + 1) & 1023;                                       \
   } while (0)

#else  /* VMX86_DEBUG */
#define ASSERT_NOT_TESTED(_c)
#define NOT_TESTED()
#define NOT_TESTED_ONCE()
#define NOT_TESTED_1024()

#endif /* VMX86_DEBUG */


#define ASSERT_NO_INTERRUPTS()   ASSERT(!INTERRUPTS_ENABLED())
#define ASSERT_HAS_INTERRUPTS()  ASSERT(INTERRUPTS_ENABLED())

/*
 * The following are only present for the OBJDIR=obj build (development).
 */
#if defined(VMX86_DEVEL) && defined(VMX86_DEBUG)
#define ASSERT_DEVEL(_c) \
   _ASSERT_PANIC(0, _c, Assert_AssertFailed, PanicAssert)

#define MEMORY_LEAK(_ptr) {                  \
   Log("Memory leak 0x%8X at %s:%d\n",       \
       _ptr,FILECODESTRING,__LINE__);        \
}

#else /* VMX86_DEBUG  && VMX86_DEVEL */

#define ASSERT_DEVEL(_c)
#define MEMORY_LEAK()
#endif /* !VMX86_DEBUG || !VMX86_DEVEL */

/* Convenience macros and definitions. Can often be used instead of #ifdef. */
#ifdef VMX86_DEBUG
#undef DEBUG_ONLY
#define vmx86_debug   1
#define DEBUG_ONLY(x) x
#else
#undef DEBUG_ONLY
#define vmx86_debug   0
#define DEBUG_ONLY(x)
#endif

#ifdef VMX86_STATS
#define vmx86_stats   1
#define STATS_ONLY(x) x
#else
#define vmx86_stats   0
#define STATS_ONLY(x)
#endif

#ifdef VMX86_DEVEL
#define vmx86_devel   1
#define DEVEL_ONLY(x) x
#else
#define vmx86_devel   0
#define DEVEL_ONLY(x)
#endif

#ifdef VMX86_LOG
#define vmx86_log     1
#define LOG_ONLY(x)   x
#else
#define vmx86_log     0
#define LOG_ONLY(x)
#endif

#ifdef VMX86_SERVER
#define vmx86_server 1
#define SERVER_ONLY(x) x
#define HOSTED_ONLY(x)
#else
#define vmx86_server 0
#define SERVER_ONLY(x)
#define HOSTED_ONLY(x) x
#endif

#ifdef VMKERNEL
#define vmkernel 1
#define VMKERNEL_ONLY(x) x
#else 
#define vmkernel 0
#define VMKERNEL_ONLY(x)
#endif

#ifdef WIN32
#define WIN32_ONLY(x) x
#define POSIX_ONLY(x)
#else
#define WIN32_ONLY(x)
#define POSIX_ONLY(x) x
#endif

#ifdef MONITOR
#define MONITOR_ONLY(x) x
#define USER_ONLY(x)
#else
#define MONITOR_ONLY(x)
#define USER_ONLY(x) x
#endif
#endif /* VM_ASSERT_H */
