/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

#ifndef _X86_H_
#define _X86_H_

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_VMM
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_DISTRIBUTE
#include "includeCheck.h"

#include "x86types.h"
#include "x86apic.h"
#include "x86desc.h"
#include "x86cpuid.h"


#define SIZE_8BIT   1
#define SIZE_16BIT  2
#define SIZE_24BIT  3
#define SIZE_32BIT  4
#define SIZE_48BIT  6
#define SIZE_64BIT  8
#define SIZE_80BIT  10
#define SIZE_128BIT 16
#define SIZE_M512   512

#define X86_MAX_INSTR_LEN  15   /* Max byte length of an x86 instruction. */

/*
 *----------------------------------------------------------------------
 *
 * OpsizeValue --
 *
 *	Called by code that can handle 32, 16, or 8 bits values to return
 *	only the defined bits for the current size.
 *
 * Results:
 *      Value of the specified bits, zero-extended out to 32 bits.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

#if defined(LOWORD) && defined(LOBYTE) // Windows doesn't have LOWORD here -- edward
static INLINE uint32
OpsizeValue(uint32 value,	// IN: The 32 bit value to be sized.
	    int opSize)		// IN: the size 
{
   ASSERT(opSize == SIZE_8BIT || opSize == SIZE_16BIT || opSize == SIZE_32BIT);

   return opSize == SIZE_8BIT  ? LOBYTE(value) : 
         (opSize == SIZE_16BIT ? LOWORD(value) : value);
}
#endif

#define NUM_IDT_VECTORS 256


/*
 * simple structures are defined but shift and masks where more
 * complicated structures use structs and bitmasks
 */


/*
 *   control registers
 */

#define CR0_PE         0x00000001
#define CR0_MP         0x00000002
#define CR0_EM         0x00000004
#define CR0_TS         0x00000008
#define CR0_ET         0x00000010
#define CR0_NE         0x00000020
#define CR0_WP         0x00010000
#define CR0_AM         0x00040000
#define CR0_NW         0x20000000
#define CR0_CD         0x40000000
#define CR0_PG         0x80000000
#define CR0_RESERVED   0x1ffaffc0

#define CR3_PWT        0x00000008
#define CR3_PCD        0x00000010
#define CR3_PDB_SHIFT  12
#define CR3_PDB_MASK   0xfffff000
#define CR3_IGNORE     0xFFF
#define PAE_CR3_IGNORE 0x1F

#define CR4_VME        0x00000001
#define CR4_PVI        0x00000002
#define CR4_TSD        0x00000004
#define CR4_DE         0x00000008
#define CR4_PSE        0x00000010
#define CR4_PAE        0x00000020
#define CR4_MCE        0x00000040
#define CR4_PGE        0x00000080
#define CR4_PCE        0x00000100

#define CR4_OSFXSR     0x00000200 // CPU/OS supports SIMD insts
#define CR4_OSXMMEXCPT 0x00000400 // #XF exception enable PIII only

#define CR4_RESERVED   0xfffff800
#define CR4_NOT_IMP    0x00000000

#define CR4_RESERVED_WO_KNI   (CR4_OSXMMEXCPT)
#define CR4_NOT_IMP_WO_KNI    (CR4_OSXMMEXCPT)

#define CR4_RESERVED_WO_VME   (CR4_VME | CR4_PVI)

#define EFLAGS_IOPL_SHIFT 12
typedef enum x86_FLAGS {
   EFLAGS_NONE         = 0,
   EFLAGS_CF           = 1 << 0,                        /* User */
   EFLAGS_SET          = 1 << 1,
   EFLAGS_PF           = 1 << 2,                        /* User */
   EFLAGS_AF           = 1 << 4,                        /* User */
   EFLAGS_ZF           = 1 << 6,                        /* User */
   EFLAGS_SF           = 1 << 7,                        /* User */
   EFLAGS_TF           = 1 << 8,                        /* Priv */
   EFLAGS_IF           = 1 << 9,                        /* Priv */
   EFLAGS_DF           = 1 << 10,                       /* User */
   EFLAGS_OF           = 1 << 11,                       /* User */
   EFLAGS_IOPL         = 3 << EFLAGS_IOPL_SHIFT,        /* Priv */
   EFLAGS_NT           = 1 << 14,                       /* Priv */
   EFLAGS_RF           = 1 << 16,                       /* Priv */
   EFLAGS_VM           = 1 << 17,                       /* Priv */
   EFLAGS_AC           = 1 << 18,                       /* Priv */
   EFLAGS_VIF          = 1 << 19,                       /* Priv */
   EFLAGS_VIP          = 1 << 20,                       /* Priv */
   EFLAGS_ID           = 1 << 21,                       /* Priv */

   EFLAGS_CLEAR        = (1 << 3) | (1 << 5) | (1 << 15) |
                         (~0u ^ ((1 << 22) - 1)), /* ~0 == (1U << 32) - 1 */
   EFLAGS_USER         = (EFLAGS_CF | EFLAGS_PF | EFLAGS_AF | EFLAGS_ZF |
                          EFLAGS_SF | EFLAGS_DF | EFLAGS_OF),
   EFLAGS_PRIV         = (EFLAGS_TF  | EFLAGS_IF  | EFLAGS_IOPL | EFLAGS_NT  |
                          EFLAGS_RF  | EFLAGS_VM  | EFLAGS_AC   | EFLAGS_VIF |
                          EFLAGS_VIP | EFLAGS_ID),
   EFLAGS_ALL          = EFLAGS_USER | EFLAGS_PRIV,

   EFLAGS_REAL_32      = (EFLAGS_ALL & ~(EFLAGS_VIP | EFLAGS_VIF | EFLAGS_VM)),
   EFLAGS_V8086_32     = (EFLAGS_ALL & ~(EFLAGS_VIP | EFLAGS_VIF |
                                         EFLAGS_VM  | EFLAGS_IOPL)),
   EFLAGS_ALL_16       = EFLAGS_ALL & 0xffff,
   EFLAGS_REAL_16      = EFLAGS_REAL_32 & 0xffff,
   EFLAGS_V8086_16     = EFLAGS_V8086_32 & 0xffff,
   EFLAGS_CLEAR_ON_EXC = (EFLAGS_TF | EFLAGS_VM | EFLAGS_RF | EFLAGS_NT),

   EFLAGS_IOPL_MASK    = EFLAGS_IOPL,
   EFLAGS__4           = 0x7fffffff,   /* ensure 4 byte encoding */
} x86_FLAGS;

#define DR6_B0         0x00000001
#define DR6_B1         0x00000002
#define DR6_B2         0x00000004
#define DR6_B3         0x00000008
#define DR6_B0123      (DR6_B0 | DR6_B1 | DR6_B2 | DR6_B3)
#define DR6_BD         0x00002000
#define DR6_BS         0x00004000
#define DR6_BT         0x00008000
#define DR6_ONES       0xffff0ff0
#define DR6_RESERVED_MASK 0xffff1ff0

#define DR7_L0         0x00000001
#define DR7_G0         0x00000002
#define DR7_L1         0x00000004
#define DR7_G1         0x00000008
#define DR7_L2         0x00000010
#define DR7_G2         0x00000020
#define DR7_L3         0x00000040
#define DR7_G3         0x00000080
#define DR7_ENABLED    0x000000ff

#define DR7_LE         0x00000100
#define DR7_GE         0x00000200
#define DR7_GD         0x00002000
#define DR7_ONES       0x00000400
#define DR7_DEFAULT    (DR7_LE | DR7_GE | DR7_ONES)

#define DR7_RW(_r,_n)  (((_r) >> (16+(_n)*4)) & 0x3)
#define DR7_L(_r,_n)   (((_r) >> ((_n)*2)) & 1)
#define DR7_G(_r,_n)   (((_r) >> (1 + (_n)*2)) & 1)
#define DR7_LEN(_r,_n) (((_r) >> (18+(_n)*4)) & 0x3)

#define DR7_RW_INST    0x0
#define DR7_RW_WRITES  0x1
#define DR7_RW_IO      0x2
#define DR7_RW_ACCESS  0x3

#define MSR_TSC              0x00000010
#define MSR_APIC_BASE        0x0000001b
#define MSR_PERFCTR0         0x000000c1
#define MSR_PERFCTR1         0x000000c2
#define MSR_MTRR_CAP         0x000000fe
#define MSR_L2CFG            0x0000011e
#define MSR_SYSENTER_CS      0x00000174
#define MSR_SYSENTER_ESP     0x00000175
#define MSR_SYSENTER_EIP     0x00000176
#define MSR_MCG_CAP          0x00000179
#define MSR_MCG_STATUS       0x0000017a
#define MSR_MCG_CTL          0x0000017b
#define MSR_EVNTSEL0         0x00000186
#define MSR_EVNTSEL1         0x00000187
#define MSR_DEBUGCTL         0x000001d9

#define MSR_MTRR_BASE0        0x00000200
#define MSR_MTRR_MASK0        0x00000201
#define MSR_MTRR_BASE1        0x00000202
#define MSR_MTRR_MASK1        0x00000203
#define MSR_MTRR_BASE2        0x00000204
#define MSR_MTRR_MASK2        0x00000205
#define MSR_MTRR_BASE3        0x00000206
#define MSR_MTRR_MASK3        0x00000207
#define MSR_MTRR_BASE4        0x00000208
#define MSR_MTRR_MASK4        0x00000209
#define MSR_MTRR_BASE5        0x0000020a
#define MSR_MTRR_MASK5        0x0000020b
#define MSR_MTRR_BASE6        0x0000020c
#define MSR_MTRR_MASK6        0x0000020d
#define MSR_MTRR_BASE7        0x0000020e
#define MSR_MTRR_MASK7        0x0000020f
#define MSR_MTRR_FIX64K_00000 0x00000250
#define MSR_MTRR_FIX16K_80000 0x00000258
#define MSR_MTRR_FIX16K_A0000 0x00000259
#define MSR_MTRR_FIX4K_C0000  0x00000268
#define MSR_MTRR_FIX4K_C8000  0x00000269
#define MSR_MTRR_FIX4K_D0000  0x0000026a
#define MSR_MTRR_FIX4K_D8000  0x0000026b
#define MSR_MTRR_FIX4K_E0000  0x0000026c
#define MSR_MTRR_FIX4K_E8000  0x0000026d
#define MSR_MTRR_FIX4K_F0000  0x0000026e
#define MSR_MTRR_FIX4K_F8000  0x0000026f
#define MSR_MTRR_DEF_TYPE     0x000002ff

#define MSR_MC0_CTL          0x00000400
#define MSR_MC0_STATUS       0x00000401
#define MSR_MC0_ADDR         0x00000402
#define MSR_MC0_MISC         0x00000403

#define MSR_LASTBRANCHFROMIP 0x000001db // Intel P6 Family
#define MSR_LASTBRANCHTOIP   0x000001dc // Intel P6 Family
#define MSR_LASTINTFROMIP    0x000001dd // Intel P6 Family
#define MSR_LASTINTTOIP      0x000001de // Intel P6 Family

#define MSR_LER_FROM_LIP     0x000001d7 // Intel Pentium4 Family
#define MSR_LER_TO_LIP       0x000001d8 // Intel Pentium4 Family
#define MSR_LASTBRANCH_TOS   0x000001da // Intel Pentium4 Family
#define MSR_LASTBRANCH_0     0x000001db // Intel Pentium4 Family
#define MSR_LASTBRANCH_1     0x000001dc // Intel Pentium4 Family
#define MSR_LASTBRANCH_2     0x000001dd // Intel Pentium4 Family
#define MSR_LASTBRANCH_3     0x000001de // Intel Pentium4 Family

/* DebugCtlMSTR bits */
#define MSR_DEBUGCTL_LBR     0x00000001

/* This ifndef is necessary because this is defined by some kernel headers. */
#ifndef MSR_K7_HWCR
#define MSR_K7_HWCR          0xc0010015 // AMD Athlon Family
#endif

/*
 * MTRR bit description
 */
#define MTRR_CAP_WC           0x400
#define MTRR_CAP_FIX          0x100
#define MTRR_CAP_VCNT_MASK    0xff

#define MTRR_DEF_ENABLE       0x800
#define MTRR_DEF_FIXED_ENABLE 0x400
#define MTRR_DEF_TYPE_MASK    0xff

#define MTRR_BASE_TYPE_MASK   0xff
#define MTRR_BASE_ADDR_MASK   (0xffffff000LL)

#define MTRR_MASK_VALID       0x800
#define MTRR_MASK_ADDR_MASK   (0xffffff000LL)

#define MTRR_TYPE_UC          0
#define MTRR_TYPE_WC          1
#define MTRR_TYPE_WT          4
#define MTRR_TYPE_WP          5
#define MTRR_TYPE_WB          6

typedef uint32 CReg;


/*
 *   segment selectors
 */

typedef uint16 Selector;

#define SELECTOR_GDT             0
#define SELECTOR_LDT             1
#define SELECTOR_RPL_SHIFT       0
#define SELECTOR_RPL_MASK        0x03u
#define SELECTOR_TI_SHIFT        2
#define SELECTOR_TI_MASK         0x4
#define SELECTOR_INDEX_SHIFT     3
#define SELECTOR_INDEX_MASK      0xfff8

#define SELECTOR_RPL(_sel)       (((Selector)(_sel)) & SELECTOR_RPL_MASK)
#define SELECTOR_TABLE(_sel)     ((((Selector)(_sel)) & SELECTOR_TI_MASK) >> SELECTOR_TI_SHIFT)
#define SELECTOR_INDEX(_sel)     (((Selector)(_sel)) >> SELECTOR_INDEX_SHIFT)
#define SELECTOR_CLEAR_RPL(_sel) (((Selector)(_sel)) & ~SELECTOR_RPL_MASK)
#define NULL_SELECTOR(_sel)      (!SELECTOR_CLEAR_RPL(_sel))

/*
 *   tasks
 */

#ifdef __GNUC__
struct Task {
#else
typedef struct Task {
#endif
   uint16     prevTask,  __prevTasku;
   uint32     esp0;
   uint16     ss0,  __ss0u;
   uint32     esp1;
   uint16     ss1,  __ss1u;
   uint32     esp2;
   uint16     ss2,  __ss2u;
   uint32     cr3;
   uint32     eip;
   uint32     eflags;
   uint32     eax;
   uint32     ecx;
   uint32     edx;
   uint32     ebx;
   uint32     esp;
   uint32     ebp;
   uint32     esi;
   uint32     edi;
   uint16     es,  __esu;
   uint16     cs,  __csu;
   uint16     ss,  __ssu;
   uint16     ds,  __dsu;
   uint16     fs,  __fsu;
   uint16     gs,  __gsu;
   uint16     ldt,  __ldtu;
   uint16     trap;
   uint16     IOMapBase;
#ifdef __GNUC__
}  __attribute__ ((packed));

typedef struct Task Task;
#else
} Task;
#endif


typedef struct {
   uint16     prevTask;
   uint16     sp0;  // static.  Unmarked fields are dynamic
   uint16     ss0;  // static
   uint16     sp1;  // static
   uint16     ss1;  // static
   uint16     sp2;  // static
   uint16     ss2;  // static 
   uint16     ip;
   uint16     flags;
   uint16     ax;
   uint16     cx;
   uint16     dx;
   uint16     bx;
   uint16     sp;
   uint16     bp;
   uint16     si;
   uint16     di;
   uint16     es;
   uint16     cs;
   uint16     ss;
   uint16     ds;
   uint16     ldt;  // static
} Task16;


/*
 *   far pointers
 */

typedef struct {
   uint32 va;
   Selector seg;
} FarPtr;

typedef struct FarPtr16 {
   uint16   offset;
   uint16   selector;
} FarPtr16;

typedef struct FarPtr32 {
   uint32   offset;
   uint16   selector;
} FarPtr32;

/*
 * General-purpose registers and segment registers 
 *
 * REG_NULL indicates the absence of a register in certain APIs. 
 * REG_ZERO denotes a special register that always contains zero.
 * In other words: 
 *    - VC->R[REG_NULL] is an error
 *    - VC->R[REG_ZERO] evaluates to zero.
 */
typedef enum RegisterName {
   REG_NULL = -1, // -1 so that NUM_REGS doesn't include REG_NULL
   REG_EAX,
   REG_ECX,
   REG_EDX,
   REG_EBX,
   REG_ESP,
   REG_EBP,
   REG_ESI,
   REG_EDI,
   REG_ZERO = 8,  /* Not a real register, not counted in "NUM_REGS". */
   NUM_REGS = 8
} RegisterName;


/* Return TRUE iff "regName" is a valid register number. */
static INLINE Bool 
x86_IsGoodRegister(RegisterName regName) 
{ 
   return (REG_NULL < regName && regName < NUM_REGS); 
}

typedef enum RegisterName8 {
   REG8_NULL = -1,
   REG8_AL,
   REG8_CL,
   REG8_DL,
   REG8_BL,
   REG8_AH,
   REG8_CH,
   REG8_DH,
   REG8_BH,
   NUM_REG8S
} RegisterName8;

/* Return TRUE iff "regName8" is a valid 8 bit register number. */
static INLINE Bool 
x86_IsGoodRegister8(RegisterName8 regName8) 
{ 
   return (REG8_NULL < regName8 && regName8 < NUM_REG8S); 
}

/* Mapping: EAX -> AL, ECX -> CL, EDX -> DL, EBX -> BL. */
static INLINE RegisterName8
x86_RegToReg8L(RegisterName reg)
{
   ASSERT(reg == REG_EAX || reg == REG_ECX ||
          reg == REG_EDX || reg == REG_EBX);

   return (RegisterName8)reg;
}

/* Mapping: EAX -> AH, ECX -> CH, EDX -> DH, EBX -> BH. */
static INLINE RegisterName8
x86_RegToReg8H(RegisterName reg)
{
   ASSERT(reg == REG_EAX || reg == REG_ECX ||
          reg == REG_EDX || reg == REG_EBX);

   return (RegisterName8)(reg + 4);
}

typedef enum SegmentName {
   SEG_NULL = -1, // -1 so that NUM_SEGS doesn't include SEG_NULL
   SEG_ES,
   SEG_CS,
   SEG_SS,
   SEG_DS,
   SEG_FS,
   SEG_GS,
   SEG_LDTR,
   SEG_TR,
   NUM_SEGS
} SegmentName;

#define NUM_NORMAL_SEGS  6

static INLINE Bool 
x86_IsGoodSegmentName(SegmentName segName)
{ 
   return SEG_NULL < segName && segName < NUM_SEGS; 
}

/* Return FALSE iff "segName" refers to the special SEG_LDTR or SEG_TR. */
static INLINE Bool 
x86_IsNormalSegmentName(SegmentName segName)
{
   ASSERT(x86_IsGoodSegmentName(segName));

   return segName < NUM_NORMAL_SEGS;
}

/* Return TRUE iff "segName" is one of SEG_ES, SEG_DS, SEG_FS, or SEG_GS. */
static INLINE Bool
x86_IsVanillaSegmentName(SegmentName segName)
{
   ASSERT(x86_IsGoodSegmentName(segName));

   return segName == SEG_DS || segName == SEG_ES ||
          segName == SEG_FS || segName == SEG_GS || segName == SEG_CS;
}

/* Bit vector representation of a set of general-purpose registers. */
typedef uint8 RegisterSet;

#define GPR_REG(reg)  ((RegisterSet)(1 << (reg)))
#define GPR_EMPTY     ((RegisterSet)0)
#define GPR_ALL       (~GPR_EMPTY)
#define GPR_EAX       GPR_REG(REG_EAX)
#define GPR_ECX       GPR_REG(REG_ECX)
#define GPR_EDX       GPR_REG(REG_EDX)
#define GPR_EBX       GPR_REG(REG_EBX)
#define GPR_ESP       GPR_REG(REG_ESP)
#define GPR_EBP       GPR_REG(REG_EBP)
#define GPR_ESI       GPR_REG(REG_ESI)
#define GPR_EDI       GPR_REG(REG_EDI)

/* Return TRUE iff the register "reg" is a member of the set "regs". */
static INLINE Bool 
GPR_IncludesReg(RegisterSet regs, RegisterName reg)
{
   return (regs & GPR_REG(reg)) != 0;
}


typedef enum ExcSource {
   IDT_FROM_UNKNOWN = -1,
   IDT_FROM_MONITOR,
   IDT_FROM_LONGJMP,
   IDT_FROM_BT,
   IDT_FROM_DIRECT_EXEC,
   IDT_NUM_EXC_SOURCES
} ExcSource;

static INLINE Bool
x86_MonitorOrLongJmpExcSource(ExcSource excSrc)
{
   ASSERT(IDT_FROM_MONITOR == 0 && IDT_FROM_LONGJMP == 1);
   return excSrc <= IDT_FROM_LONGJMP;
}

static INLINE Bool
x86_IsGoodExcSource(ExcSource excSrc)
{
    ASSERT(excSrc >= IDT_FROM_UNKNOWN && excSrc < IDT_NUM_EXC_SOURCES);
    return excSrc != IDT_FROM_UNKNOWN;
}

/*
 * Exception frame, as used by the monitor.
 */
typedef struct ExcFrame {
   // pushed by sw
   ExcSource   source;

   uint16      es, __esu;
   uint16      csCopy, __csCopyu; // Not restored, usually not valid.
   uint16      ssCopy, __ssCopyu; // Not restored, usually not valid.
   uint16      ds, __dsu;
   uint16      fs, __fsu;
   uint16      gs, __gsu;

   Reg         eax;
   Reg         ecx;
   Reg         edx;
   Reg         ebx;
   Reg         espCopy;           // Not restored, usually not valid.
   Reg         ebp;
   Reg         esi;
   Reg         edi;

   // pushed by SW or HW depending on the interrupt
   Reg         errorCode;
   
   // pushed by HW
   Reg         eip;
   uint16      cs, __csu;
   x86_FLAGS   eflags;
   
   // pushed by HW on protection level changes
   Reg         esp;
   uint16      ss, __ssu;

   // pushed by HW coming from virtual 8086 mode
   uint16      v8086_es, __v8086_esu;
   uint16      v8086_ds, __v8086_dsu;
   uint16      v8086_fs, __v8086_fsu;
   uint16      v8086_gs, __v8086_gsu;
} ExcFrame;

/*
 * Can be used to treat the segment selectors in ExcFrame as an array.
 */
typedef struct PaddedSelector {
   uint16 sel;
   uint16 pad;
} PaddedSelector;

/*
 *----------------------------------------------------------------------
 * ExcFrameToSelectorArray --
 *
 *    Returns a pointer to an array of PaddedSelectors allowing
 *    clients to iterate over the segment registers.  The values for
 *    CS and SS are copied into the array and must be passed in
 *    because their locations differ between exceptions which change
 *    privilege level and those that do not.
 *----------------------------------------------------------------------
 */
static INLINE PaddedSelector*
ExcFrameToSelectorArray(ExcFrame *excFrame, Selector cs, Selector ss)
{
   PaddedSelector *ps = (PaddedSelector*) &excFrame->es;
   ASSERT(&ps[SEG_ES].sel == (Selector*) &excFrame->es &&
          &ps[SEG_CS].sel == (Selector*) &excFrame->csCopy &&
          &ps[SEG_SS].sel == (Selector*) &excFrame->ssCopy &&
          &ps[SEG_DS].sel == (Selector*) &excFrame->ds &&
          &ps[SEG_FS].sel == (Selector*) &excFrame->fs &&
          &ps[SEG_GS].sel == (Selector*) &excFrame->gs);
   ps[SEG_CS].sel = cs;
   ps[SEG_SS].sel = ss;
   return ps;
}

/*
 * X86-defined stack layouts for interrupts, exceptions, irets, calls, etc.
 */

/*
 * Layout of the 32-bit stack frame on exception.
 */
typedef struct x86ExcFrame32 {
   uint32         eip;
   union {
      struct {
         uint16   sel, __sel;
      }           cs16;
      uint32      cs32;
   } u;
   uint32         eflags;
} x86ExcFrame32;


/*
 * Layout of the 32-bit stack frame on inter-level transfer.
 */
typedef struct x86ExcFrame32IL {
   uint32      errorCode;
   uint32      eip;
   uint16      cs, __csu;
   uint32      eflags;
   uint32      esp;
   uint16      ss, __ssu;
} x86ExcFrame32IL;


/*
 * Layout of the 16-bit stack frame on exception.
 */
typedef struct x86ExcFrame16 {
   uint16   eip;
   uint16   cs;
   uint16   eflags;
} x86ExcFrame16;

/*
 * Layout of the 32-bit stack frame on exception
 * from V8086 mode. It is also a superset
 * of inter-level exception stack frame, which
 * in turn is superset of intra-level exception
 * stack frame.
 */
typedef struct x86ExcFrameV8086 {
   uint32         eip;
   union {
      struct {
         uint16   sel, __sel;
      }           cs16;
      uint32      cs32;
   } u;
   uint32         eflags;
   uint32         esp;
   uint16         ss, __ss;
   uint16         es, __es;
   uint16         ds, __ds;
   uint16         fs, __fs;
   uint16         gs, __gs;
} x86ExcFrameV8086;

/*
 * Layout of the 32-bit stack frame on exception
 * from V8086 mode with errorCode. It is
 * superset of SegmentExcFrameV8086.
 */
typedef struct x86ExcFrameV8086WithErrorCode {
   uint32         errorCode;
   uint32         eip;
   union {
      struct {
         uint16   sel, __sel;
      }           cs16;
      uint32      cs32;
   } u;
   uint32         eflags;
   uint32         esp;
   uint16         ss, __ss;
   uint16         es, __es;
   uint16         ds, __ds;
   uint16         fs, __fs;
   uint16         gs, __gs;
} x86ExcFrameV8086WithErrorCode;

/*
 * Layout of the stack on a 32 bit far call.
 */
typedef struct x86CallStack32 {
   uint32   eip;
   uint16   cs, __cs;
} x86CallStack32;

/*
 * Layout of the stack on a 16 bit far call.
 */
typedef struct x86CallStack16 {
   uint16   ip;
   uint16   cs;
} x86CallStack16;

/*
 * Layout of the stack on a 32 bit far call.
 */
typedef struct x86CallGateStack32 {
   uint32   eip;
   uint16   cs, __cs;
   uint32   esp;
   uint16   ss, __ss;
} x86CallGateStack32;

/*
 * Layout of the stack on a 16 bit far call.
 */
typedef struct x86CallGateStack16 {
   uint16   ip;
   uint16   cs;
   uint16   sp;
   uint16   ss;
} x86CallGateStack16;


/*
 *   page tables
 */

#define PTE_P           0x00000001
#define PTE_RW          0x00000002
#define PTE_US          0x00000004
#define PTE_PWT         0x00000008
#define PTE_PCD         0x00000010
#define PTE_A           0x00000020
#define PTE_D           0x00000040
#define PTE_PS          0x00000080
#define PTE_G           0x00000100

#define PTE_FLAGS       (PTE_P|PTE_RW|PTE_US|PTE_G)

#define PTE_PAGE_TABLE  (PTE_P|PTE_RW|PTE_US|PTE_A|PTE_D) 

/*
 * This is usable for 4M or 2M pde's (both PAE and non-PAE). Use the PTE
 * macros for large pages.
 */ 
#define PDE_PAGE_TABLE  (PTE_P | PTE_RW | PTE_US | PTE_A) 

#define PTE_PRESENT(_pte)   (((_pte) & PTE_P) != 0)
#define PTE_WRITEABLE(_pte) (((_pte) & PTE_RW) != 0)
#define PTE_DIRTY(_pte)     (((_pte) & PTE_D) != 0)
#define PTE_USER(_pte)      (((_pte) & PTE_US) != 0)
#define PTE_GLOBAL(_pte)    (((_pte) & PTE_G) != 0)
#define PTE_LARGEPAGE(_pte) (((_pte) & PTE_PS) != 0)

#define PTE_AVAIL_MASK       0xe00
#define PTE_AVAIL_SHIFT      9
#define PTE_PFN_MASK         0xfffff000
#define PAE_PTE_PFN_MASK     CONST64U(0xffffff000)
#define PTE_PFN_SHIFT        12
#define PAE_PTE_PFN_SHIFT    12
#define PTE_2_PFN(_pte)      (((_pte) & PTE_PFN_MASK) >> PTE_PFN_SHIFT)
#define PAE_PTE_2_PFN(_pte)  (((_pte) & PAE_PTE_PFN_MASK) >> PAE_PTE_PFN_SHIFT)
#define PTE_AVAIL(_pte)      (((_pte) & PTE_AVAIL_MASK) >> PTE_AVAIL_SHIFT)

#define PDE_PFN_MASK         0xffc00000
#define PAE_PDE_PFN_MASK     0xfffe00000LL
#define PDE_PFN_SHIFT        PTE_PFN_SHIFT
#define PAE_PDE_PFN_SHIFT    PAE_PTE_PFN_SHIFT

/* Use these only for 4M or 2M pde's, not the 4K ones. */
#define PDE_2_PFN(_pte)      (((_pte) & PDE_PFN_MASK) >> PTE_PFN_SHIFT)
#define PAE_PDE_2_PFN(_pde)  (((_pde) & PAE_PDE_PFN_MASK) >> PAE_PDE_PFN_SHIFT)

#define PDPTE_PFN_MASK       0xffffff000LL
#define PDPTE_PFN_SHIFT      12
#define PDPTE_2_PFN(_p)      (((_p) & PDPTE_PFN_MASK) >> PDPTE_PFN_SHIFT)
#define PDPTE_RESERVED       0xFFFFFFF0000001E6ULL
#define PDPTE_INVALID        0x6 /* Reserved bits */

#define NUM_PAE_PDIRS        4
#define PDES_PER_PDIR        (PAGE_SIZE/sizeof(VM_PDE))
#define PTES_PER_PDIR        PDES_PER_PDIR
#define PAE_PDES_PER_PDIR    (PAGE_SIZE/sizeof(VM_PAE_PDE))
#define PTES_PER_PGTBL       (PAGE_SIZE/sizeof(VM_PTE))
#define PAE_PTES_PER_PGTBL   (PAGE_SIZE/sizeof(VM_PAE_PTE))

#define PGOFF_MASK            0x3ff
#define PAE_PGOFF_MASK        0x1ff
#define PDOFF_MASK            0x3ff
#define PAE_PDOFF_MASK        0x1ff
#define PDPTOFF_MASK          0x3
#define PAE_RANGE_MASK        0x7ff

#define PDE_4M_RSVD_MASK      0x003e0000
#define PAE_PTE_4K_RSVD_MASK  0xfffffff000000000LL
#define PAE_PDE_4K_RSVD_MASK  0xfffffff000000000LL
#define PAE_PDE_2M_RSVD_MASK  0xfffffff0001fe000LL

#define PFN_2_PDPTOFF(_a)     (((_a) >> 18) & PDPTOFF_MASK)
#define PFN_2_PDOFF(_a)       (((_a) >> 10) & PDOFF_MASK)
#define PAE_PFN_2_PDOFF(_a)   (((_a) >> 9) & PAE_PDOFF_MASK)
#define PFN_2_PGOFF(_a)       ((_a) & PGOFF_MASK)
#define PAE_PFN_2_PGOFF(_a)   ((_a) & PAE_PGOFF_MASK)
#define PFN_2_RANGE(_pfn)     PFN_2_PDOFF(_pfn)
#define PAE_PFN_2_RANGE(_pfn) (((_pfn) >> 9) & PAE_RANGE_MASK)

#define RANGE_2_LA(_r)        ((_r) << 22)
#define PAE_RANGE_2_LA(_r)    ((_r) << 21)
#define RANGE_2_PDPTOFF(_r)   (((_r) >> 9) & 0x3)
#define RANGE_2_PDOFF(_r)     ((_r))
#define PAE_RANGE_2_PDOFF(_r) ((_r) & 0x1ff)

#define LA_2_PDPTOFF(_a)      (((_a) >> 30) & PDPTOFF_MASK)
#define LA_2_PDOFF(_a)        (((_a) >> 22) & PDOFF_MASK)
#define PAE_LA_2_PDOFF(_a)    (((_a) >> 21) & PAE_PDOFF_MASK)
#define LA_2_PGOFF(_a)        (PFN_2_PGOFF(VA_2_VPN(_a)))
#define PAE_LA_2_PGOFF(_a)    (PAE_PFN_2_PGOFF(VA_2_VPN(_a)))
#define LA_2_RANGE(_a)        LA_2_PDOFF((_a))
#define PAE_LA_2_RANGE(_a)    (((_a) >> 21) & PAE_RANGE_MASK)
#define PAE_LA_2_PDPTOFF(_a)  (((_a) >> 30) & 0x3)

/* This following macro only work for flat model were VA==LA */
#define VA_2_PDOFF(_a)        ((_a) >> 22)
#define VA_2_PGOFF(_a)        (PFN_2_PGOFF(VA_2_VPN(_a)))

#define PTE_2_PA(_a)          (PPN_2_PA(PTE_2_PFN(_a)))
#define PAE_PTE_2_PA(_a)      (PPN_2_PA(PAE_PTE_2_PFN(_a)))
#define PDE_2_PA(_a)          (PPN_2_PA(PDE_2_PFN(_a)))
#define PAE_PDE_2_PA(_a)      (PPN_2_PA(PAE_PDE_2_PFN(_a)))

#define PDE_FLAGS(_pde)       ((_pde) & 0xfff)  
#define PAE_PDE_FLAGS(_pde)   ((_pde) & 0xfff)  
/*
 *   exception error codes
 */

#define EXC_NONE         -1
#define EXC_DE            0
#define EXC_DB            1
#define EXC_NMI           2
#define EXC_BP            3
#define EXC_OF            4
#define EXC_BR            5
#define EXC_UD            6
#define EXC_NM            7
#define EXC_DF            8
#define EXC_TS           10
#define EXC_NP           11
#define EXC_SS           12
#define EXC_GP           13
#define EXC_PF           14
#define EXC_MF           16
#define EXC_AC           17
#define EXC_MC           18
#define EXC_XF           19  // SIMD exception

/* Error code flags */
#define PF_P            0x1
#define PF_RW           0x2
#define PF_US           0x4
#define PF_RSVD         0x8

#define ERRCODE_EXT     0x1
#define ERRCODE_IDT     0x2

/*
 *  instruction decoding
 */

#define MODRM_MOD(_m)  (((_m) >> 6) & 0x3)
#define MODRM_REG(_m)  (((_m) >> 3) & 0x7)
#define MODRM_RM(_m)   (((_m) >> 0) & 0x7)

/*
 * Debug registers
 */

#define DEBUG_STATUS_B0   (1<<0)
#define DEBUG_STATUS_B1   (1<<1)
#define DEBUG_STATUS_B2   (1<<2)
#define DEBUG_STATUS_B3   (1<<3)
#define DEBUG_STATUS_DB   (1<<13)
#define DEBUG_STATUS_BS   (1<<14)
#define DEBUG_STATUS_BT   (1<<15)

typedef struct DebugControlRegister {

   int l0:1;
   int g0:1;
   int l1:1;
   int g1:1;
   int l2:1;
   int g2:1;
   int l3:1;
   int g3:1;
   
   int le:1;
   int ge:1;
   int oo1:3;
   
   int gd:1;
   int oo:2;
   
   int rw0:2;
   int len0:2;
   int rw1:2;
   int len1:2;
   int rw2:2;
   int len2:2;
   int rw3:2;
   int len3:2;
   
} DebugControlRegister;

/*
 * Macros for PL4 page table used by x86_64.
 */
#define PL4_PML4OFF_MASK      0x1ff
#define PL4_PDPTOFF_MASK      PL4_PML4OFF_MASK
#define PL4_PDOFF_MASK        PL4_PML4OFF_MASK
#define PL4_PGOFF_MASK        PL4_PML4OFF_MASK

#define PL4_PFN_2_PML4OFF(_a) (((_a) >> 27) & PL4_PML4OFF_MASK)
#define PL4_PFN_2_PDPTOFF(_a) (((_a) >> 18) & PL4_PDPTOFF_MASK)
#define PL4_PFN_2_PDOFF(_a)   (((_a) >> 9) & PL4_PDOFF_MASK)
#define PL4_PFN_2_PGOFF(_a)   ((_a) & PL4_PDPTOFF_MASK)

#define PL4_PTE_SHIFT         12
#define PL4_A1_SHIFT          51
#define PL4_A2_SHIFT          9

#define PL4_A1_MASK           CONST64(0x7ff)
#define PL4_A2_MASK           CONST64(0x7)
#define PL4_FLAGS_MASK        CONST64(0x80000000000001ff)
#define PL4_PDE_MASK          CONST64(0xffffffffffffff7f)
#define PL4_CR3_FLAGS_MASK    CONST64(0x18)
#define PL4_PTE_NX            CONST64(0x8000000000000000)

#define PL4_MAKE_CR3(_mpfn, _flags) \
                              (((uint64)(_mpfn) << PL4_PTE_SHIFT) | \
                               ((_flags) & PL4_CR3_FLAGS_MASK))

#define PL4_MAKE_PTE(_mpfn, _a1, _a2, _flags) \
                              (((uint64)(_mpfn) << PL4_PTE_SHIFT) | \
                               ((((uint64)_a1) & PL4_A1_MASK) << PL4_A1_SHIFT)\
                               | (((_a2) & PL4_A2_MASK) << PL4_A2_SHIFT) \
                               | ((uint64)(_flags) & PL4_FLAGS_MASK))

#define PL4_MAKE_PML4E(_mpfn, _a1, _a2, _flags) \
                              PL4_MAKE_PTE(_mpfn, _a1, _a2, _flags)

#define PL4_MAKE_PDPE(_mpfn, _a1, _a2, _flags) \
                              PL4_MAKE_PTE(_mpfn, _a1, _a2, _flags)

#define PL4_MAKE_PDE(_mpfn, _a1, _a2, _flags) \
                             (PL4_MAKE_PTE(_mpfn, _a1, _a2, _flags) & \
                              PL4_PDE_MASK)

                 
#define MAKE_SELECTOR(_index, _ti, _RPL)                     \
( (((_index) << SELECTOR_INDEX_SHIFT) & SELECTOR_INDEX_MASK) \
  | (((_ti)  << SELECTOR_TI_SHIFT)    & SELECTOR_TI_MASK)    \
  | (((_RPL) << SELECTOR_RPL_SHIFT)   & SELECTOR_RPL_MASK) )


#define MAKE_PTE(_mpfn, _avail, _flags) \
                        (((_mpfn) << PTE_PFN_SHIFT) \
                         | (((_avail) << PTE_AVAIL_SHIFT) & PTE_AVAIL_MASK) \
                         | (_flags))

#define PAE_MAKE_PTE(_mpfn, _avail, _flags) \
     ((((VM_PAE_PTE) (_mpfn)) << PAE_PTE_PFN_SHIFT) \
      | (((_avail) << PTE_AVAIL_SHIFT) & PTE_AVAIL_MASK) \
      | (_flags))

#define MAKE_PDE(_mpfn, _avail, _flags) MAKE_PTE(_mpfn, _avail, _flags)
#define PAE_MAKE_PDE(_mpfn, _avail, _flags) PAE_MAKE_PTE(_mpfn, _avail, _flags)
#define MAKE_PDPTE(_mpfn, _avail, _flags) PAE_MAKE_PTE(_mpfn, _avail, _flags)
#define PDPTE_FLAGS (PTE_P)
     


/*
 * SMM State Save Map 
 * See the Intel processor manual for documentation.
 * The extract layout of the structure is guesswork.  It should really
 * matter also as the Intel documented fields are at the right offset. 
 */

/*
 * SMMSegState - Store the "hidden" state of a segment descriptor in 
 * the SMM save state map. NOTE: Format is somewhat defined by the 
 * Intel hardware.
 */
typedef struct SMMSegState {
   uint32 limit;   /* Limit from descriptor */
   uint32 base;    /* Base from descriptor */
   uint32 rights;  /* Rest of the descriptor */
} SMMSegState;

/*
 * Format the the SMM save state map.
 * NOTE: This too is defined by the Intel hardware.
 */

typedef struct SMMStateSaveMap {
   union {
      uint32 selState;       // We steal a enough of this (hopefully) unused
                             // space to stash away the selector state. 
      uint8  reserved0[248];
   } u;
   uint32 smbase;
   uint32 revId; 
   uint16 ioInstRestart;
   uint16 haltRestart;
   uint8  reserved01[12];
   uint32 ioInstEIP; 
   uint8  reserved1[18];
   uint32 cr4reg;
   uint8  reserved2[4];
   SMMSegState hiddenSegPart[7];
   SMMSegState gdtr;
   SMMSegState idtr;
   SMMSegState hiddenTRpart;
   uint32 segRegs[8];
   Reg       dr7reg;
   Reg       dr6reg;
   Reg       gpRegs[8];
   Reg       eip;
   x86_FLAGS eflags;
   Reg       cr3reg;
   Reg       cr0reg;
} SMMStateSaveMap;

#define SMM_SAVEMAP_SIZE   512
#define SMRAM_DEFAULT_BASE 0x30000
#define SMRAM_ENTRY_POINT 0x8000
#define SMRAM_SAVE_STATE_OFFSET 0xfe00

#define SMM_RIGHTS_BITS   0x00f0ff00

/*
 * FPU FSAVE/FXSAVE area format. 
 *
 * fsave_hard_struct - Memory format of the FSAVE/FNSAVE instruction. 108 bytes
 * fxsave_hard_struct - Memory format of the FXSAVE instruction. 512 bytes
 *                    Note: fxsave_hard_struct must be 128bit aligned. 
 *
 * Taken from linux headers
 */

struct fsave_hard_struct {
   long   cwd;
   long   swd;
   long   twd;
   long   fip;
   long   fcs;
   long   foo;
   long   fos;
   long   st_space[20];   /* 8*10 bytes for each FP-reg = 80 bytes */
};

struct fxsave_hard_struct {
   unsigned short fxcwd;
   unsigned short fxswd;
   unsigned short fxtwd;
   unsigned short fxfopcode;
   long     fxfip;
   short    fxfcs;
   short    __reserved_00;
   long     fxfoo;
   short    fxfos;
   short    __reserved_01;
   long     mxcsr;
   long     __reserved_02;
   long     st_space[32];     /* 8*16 bytes for each FP/MMX-reg = 128 bytes */
   long     xmm_space[32];    /* 8*16 bytes for each XMM-reg = 128 bytes */
   long     __reserved_03 [14*4]; /* 14 16byte lines for remainder */
};

/* 
 *---------------------------------------------------------------------------
 * IsGoodMPN --
 *    Return TRUE if "mpn" looks plausible. We could make this
 *    test stricter in the non-PAE case. We could even take into
 *    account the amount of memory on the host.
 *---------------------------------------------------------------------------
 */
static INLINE Bool IsGoodMPN(MPN mpn) { return (mpn & 0xff000000) == 0; }


/*
 *----------------------------------------------------------------------
 *
 * x86_ErrorCode2Offset --
 *
 *     Transforms an error code into a byte offset by masking off
 *     the top 16 bits and the bottom three bits.  Note that
 *     some exceptions don't push any error codes, some always push
 *     zero, and the page fault exception uses its own unique error
 *     code format that does not make reference to a selector/offset.
 *
 * Results:
 *     Returns the descriptor table offset referenced by the error code.
 *
 * Side effects:
 *     None.
 *
 *----------------------------------------------------------------------
 */

static INLINE int
x86_ErrorCode2Offset(uint32 errorCode)
{
   return errorCode & ~0xffff0007;
}


/*
 *----------------------------------------------------------------------
 *
 * x86_ErrorCode2RPL0Selector --
 *
 *     Returns the RPL0 version of the selector referenced by an error
 *     code.  Note that it's impossible to reconstruct the RPL of the
 *     original selector that caused the error code to be generated,
 *     because the RPL bits of the selector are used as flag bits
 *     in the error code.  It is not legal to call this function on
 *     an error code with the IDT bit set, since such error codes
 *     refer to IDT offsets, not GDT/LDT selectors.
 *
 * Results:
 *     Returns the descriptor table offset referenced by the error code.
 *
 * Side effects:
 *     None.
 *
 *----------------------------------------------------------------------
 */

static INLINE Selector
x86_ErrorCode2RPL0Selector(uint32 errorCode)
{
   ASSERT((errorCode & ERRCODE_IDT) == 0);
   return ((Selector) errorCode & ~0xffff0003);
}


/*
 *----------------------------------------------------------------------
 *
 * x86_ExcUsesErrorCode --
 *
 *     Determines whether the specified fault number pushes an error code
 *     on the stack when the vector is invoked.
 *
 * Results:
 *     TRUE if the exception number uses an error code, and FALSE otherwise.
 *
 * Side effects:
 *     None.
 *
 *----------------------------------------------------------------------
 */

static INLINE Bool
x86_ExcUsesErrorCode(int exc)
{
   return ((exc >= EXC_TS && exc <= EXC_PF) || exc == EXC_DF || exc == EXC_AC);
}


/*
 *----------------------------------------------------------------------
 *
 * x86_ExcUsesSelectorErrorCode -- 
 *
 *      Determines if a fault number is one that can push an error code
 *      that refers to a selector.  Note that some faults will sometimes
 *      explicitly push an error code of 0 for some conditions while
 *      pushing "normal" error codes for other fault conditions.
 *
 * Results:
 *      TRUE if this fault number may push an error code that
 *      refers to a segment and FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static INLINE Bool
x86_ExcUsesSelectorErrorCode(int exc)
{
   return (exc >= EXC_TS && exc <= EXC_GP);
}

#endif /* _X86_H_ */
