/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

#ifndef _X86DESC_H_
#define _X86DESC_H_

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_VMM
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_DISTRIBUTE
#include "includeCheck.h"

/*
 * Segment Descriptors.
 */

typedef struct Descriptor {
   unsigned   limit_lo  : 16;
   unsigned   base_lo   : 16;
   unsigned   base_mid  : 8;
   unsigned   type      : 4;
   unsigned   S         : 1;
   unsigned   DPL       : 2;
   unsigned   present   : 1;
   unsigned   limit_hi  : 4;
   unsigned   AVL       : 1;
   unsigned   longmode  : 1;    // only used in x86_64
   unsigned   DB        : 1;
   unsigned   gran      : 1;
   unsigned   base_hi   : 8;
} Descriptor;

typedef struct CallGate {
   unsigned   offset_lo : 16;
   unsigned   selector  : 16;
   unsigned   parms     : 5;
   unsigned   zero0     : 3;
   unsigned   type      : 5;
   unsigned   DPL       : 2;
   unsigned   present   : 1;
   unsigned   offset_hi : 16;
} CallGate;

typedef union {
   Descriptor desc;
   CallGate   cg;
   uint32     word[2];
} DescriptorUnion;

/*
 * Symbolic names for various offsets used to construct segment descriptors.
 */
/* Lower dword */
#define X86DESC_BASE_LO_SHIFT  16
#define X86DESC_LIMIT_LO_SHIFT  0
/* Upper dword */
#define X86DESC_BASE_HI_SHIFT  24
#define X86DESC_GRAN_SHIFT     23
#define X86DESC_DB_SHIFT       22
#define X86DESC_ZERO_SHIFT     21
#define X86DESC_LIMIT_HI_SHIFT 16
#define X86DESC_P_SHIFT        15
#define X86DESC_DPL_SHIFT      13
#define X86DESC_S_SHIFT        12
#define X86DESC_TYPE_SHIFT      8
#define X86DESC_BASE_MID_SHIFT  0


/* 
 * Descriptors store a 32bit segment base in 3 parts (low,
 * mid, high) and the 20bit limit in 2 parts (low, high).  The
 * following macros extract these components from the original
 * base and limit.
 */

#define BASE_LO(_dw)              ((_dw) & 0xffff)
#define BASE_MID(_dw)             (((_dw) >> 16) & 0xff)
#define BASE_HI(_dw)              (((_dw) >> 24) & 0xff)
#define LIMIT_LO(_dw)             ((_dw) & 0xffff)            /* Descriptor */
#define LIMIT_HI(_dw)             (((_dw) >> X86DESC_LIMIT_HI_SHIFT) & 0xf)
#define OFFSET_LO(_dw)            ((_dw) & 0xffff)            /* CallGate */
#define OFFSET_HI(_dw)            (((_dw) >> 16) & 0xffff)

/*
 * Accessor functions for descriptors.  
 * 
 * Note: The fields of a descriptor should always be accessed with the
 * following functions.  ANSI C specifies that any expression
 * involving integer types smaller than an int have all the variables
 * automatically promoted to a *signed* int.  This means that
 * expressions that use the fields of a descriptor directly will
 * treat them as signed quantities.  This could have unwanted effects,
 * e.g. base and limit added together could result in a negative
 * quantity.  The functions that read the bitfields always return an
 * unsigned integer so as to avoid any potential signed/unsigned
 * problems.  Functions to write the bitfields are also provided for
 * consistency.  
 */

static INLINE uint32 Desc_Type(const Descriptor *d)     { return d->type; }
static INLINE uint32 Desc_S(const Descriptor *d)        { return d->S; }
static INLINE uint32 Desc_DPL(const Descriptor *d)      { return d->DPL; }
static INLINE uint32 Desc_Present(const Descriptor *d)  { return d->present; }
static INLINE uint32 Desc_AVL(const Descriptor *d)      { return d->AVL; }
static INLINE uint32 Desc_DB(const Descriptor *d)       { return d->DB; }
static INLINE uint32 Desc_Gran(const Descriptor *d)     { return d->gran; }

static INLINE LA
Desc_GetBase(const Descriptor *d)
{
   return (d->base_hi << 24) | (d->base_mid << 16) | d->base_lo;
}

static INLINE VA32
Desc_GetLimit(const Descriptor *d)
{
   return (d->limit_hi << 16) | d->limit_lo;
}

static INLINE Bool
Desc_EqualIgnoreAccessed(const Descriptor *d1, const Descriptor *d2)
{
   const DescriptorUnion *du1 = (const DescriptorUnion*) d1;
   const DescriptorUnion *du2 = (const DescriptorUnion*) d2;
   uint32 mask = ~0u;
   if (Desc_S(d1)) {
      mask = ~(1u << 8);
   }
   return du1->word[0] == du2->word[0] &&
      (du1->word[1] & mask) == (du2->word[1] & mask);
}

static INLINE void Desc_SetType(Descriptor *d, uint32 val)     { d->type     = val; }
static INLINE void Desc_SetS(Descriptor *d, uint32 val)        { d->S        = val; }
static INLINE void Desc_SetDPL(Descriptor *d, uint32 val)      { d->DPL      = val; }
static INLINE void Desc_SetPresent(Descriptor *d, uint32 val)  { d->present  = val; }
static INLINE void Desc_SetDB(Descriptor *d, uint32 val)       { d->DB       = val; }
static INLINE void Desc_SetGran(Descriptor *d, uint32 val)     { d->gran     = val; }

static INLINE void
Desc_SetBase(Descriptor *d, LA newBase)
{
   d->base_hi  = BASE_HI(newBase);
   d->base_mid = BASE_MID(newBase);
   d->base_lo  = BASE_LO(newBase);
#ifdef ASSERT
   ASSERT(Desc_GetBase(d) == newBase);
#endif
}

static INLINE void
Desc_SetLimit(Descriptor *d, VA32 newLimit)
{
   d->limit_lo = LIMIT_LO(newLimit);
   d->limit_hi = LIMIT_HI(newLimit);
#ifdef ASSERT
   ASSERT(Desc_GetLimit(d) == newLimit);
#endif
}

static INLINE VA
CallGate_GetOffset(const CallGate *cg)
{
   return (cg->offset_hi << 16) | cg->offset_lo;
}

static INLINE void
CallGate_SetOffset(CallGate *cg, VA32 offset)
{
   cg->offset_lo = OFFSET_LO(offset);
   cg->offset_hi = OFFSET_HI(offset);
#ifdef ASSERT
   ASSERT(CallGate_GetOffset(cg) == offset);
#endif
}

/*
 *-----------------------------------------------------------------------------
 *
 * Desc_SetDescriptor --
 *
 *      Set a descriptor with the specified properties.
 *
 *      NOTE: the implementation of Desc_SetDescriptor() assumes a 
 *      little-endian byte order.  
 *
 *-----------------------------------------------------------------------------
 */
static INLINE void 
Desc_SetDescriptor(Descriptor *d, LA base, VA32 limit, uint32 type, 
                   uint32 S, uint32 DPL, uint32 present, uint32 DB, 
                   uint32 gran)
{
   DescriptorUnion *desc = (DescriptorUnion *) d;

   desc->word[0] = BASE_LO(base)   << X86DESC_BASE_LO_SHIFT  |
                   LIMIT_LO(limit) << X86DESC_LIMIT_LO_SHIFT;

   desc->word[1] = BASE_HI(base)   << X86DESC_BASE_HI_SHIFT  | 
                   gran            << X86DESC_GRAN_SHIFT     | 
                   DB              << X86DESC_DB_SHIFT       | 
                   LIMIT_HI(limit) << X86DESC_LIMIT_HI_SHIFT | 
                   present         << X86DESC_P_SHIFT        | 
                   DPL             << X86DESC_DPL_SHIFT      |
                   S               << X86DESC_S_SHIFT        | 
                   type            << X86DESC_TYPE_SHIFT     | 
                   BASE_MID(base)  << X86DESC_BASE_MID_SHIFT;

   /* 
    * Assert that all the fields were properly filled in.
    */

#ifdef ASSERT
   ASSERT(Desc_GetBase(d) == base);
   ASSERT(Desc_GetLimit(d) == limit);
   ASSERT(Desc_Type(d) == type);
   ASSERT(Desc_S(d) == S);
   ASSERT(Desc_DPL(d) == DPL);
   ASSERT(Desc_Present(d) == present);
   ASSERT(Desc_DB(d) == DB);
   ASSERT(Desc_Gran(d) == gran);
   ASSERT(d->AVL == 0);
   ASSERT(d->longmode == 0);
#endif
}


/*
 * Accessor functions that operate directly on a descriptor.  These
 * are included only for backwards compatibility with existing macros.  
 */

static INLINE uint32 DESC_TYPE(Descriptor d)       { return d.type; }
static INLINE uint32 DESC_S(Descriptor d)          { return d.S; }
static INLINE uint32 DESC_DPL(Descriptor d)        { return d.DPL; }
static INLINE uint32 DESC_PRESENT(Descriptor d)    { return d.present; }

#define DT_CODE(_d)               ( DESC_S(_d) && ((DESC_TYPE(_d) & 0x8) == 0x8))
#define DT_CONFORMING_CODE(_d)    ( DESC_S(_d) && ((DESC_TYPE(_d) & 0xc) == 0xc))
#define DT_NONCONFORMING_CODE(_d) ( DESC_S(_d) && ((DESC_TYPE(_d) & 0xc) == 0x8))
#define DT_READABLE_CODE(_d)      ( DESC_S(_d) && ((DESC_TYPE(_d) & 0xa) == 0xa))
#define DT_DATA(_d)               ( DESC_S(_d) && ((DESC_TYPE(_d) & 0x8) == 0x0))
#define DT_WRITEABLE_DATA(_d)     ( DESC_S(_d) && ((DESC_TYPE(_d) & 0xa) == 0x2))
#define DT_EXPAND_DOWN(_d)        ( DESC_S(_d) && ((DESC_TYPE(_d) & 0xc) == 0x4))
#define DT_CALL_GATE(_d)          (!DESC_S(_d) && ((DESC_TYPE(_d) & 0x7) == 0x4))
#define DT_LDT(_d)                (!DESC_S(_d) && ((DESC_TYPE(_d) & 0xf) == 0x2))
#define DT_TASK_GATE(_d)          (!DESC_S(_d) && ((DESC_TYPE(_d) & 0xf) == 0x5))
#define DT_TSS(_d)                (!DESC_S(_d) && ((DESC_TYPE(_d) & 0x5) == 0x1))
#define DT_AVAIL_TSS(_d)          (!DESC_S(_d) && ((DESC_TYPE(_d) & 0x7) == 0x1))

#define DT_ACCESS                 0x2
#define DT_32BIT                  0x8
#define DT_TSS_BUSY               0x2

#define DATA_DESC                 0x2
#define CODE_DESC                 0xa
#define TASK_DESC                 0x9  // TSS available
#define TASK_DESC_BUSY            0xb  // TSS busy
#define TASK16_DESC               0x1  // 16-bit TSS available
#define TASK16_DESC_BUSY          0x3  // 16-bit TSS busy


/*
 *----------------------------------------------------------------------
 * Desc_DBSize --
 *   Given descriptor, return the code/stack size that it specifies.
 *----------------------------------------------------------------------
 */
static INLINE int
Desc_DBSize(Descriptor *desc)
{
   /*
    * Code/stack size is determined by the D/B bit; bit 22 of the 2nd word.
    * Shift the bit to position 1, mask it out, add 2. Result: 2 or 4.
    */
   int cSz = ((((uint32*)desc)[1] >> 21) & 2) + 2;
   ASSERT(cSz == (Desc_DB(desc) ? 4 : 2));
   return cSz;
}


/*
 *----------------------------------------------------------------------
 *
 * Desc_ExpandedLimit --
 *
 *      Return the limit in bytes of the descriptor.  This is correct for both
 *      expand up and expand down limits.  For expand down limits, the page 
 *      corresponding to the limit << 12 is *not* included in the segment when
 *      the granularity bit is set.  This makes it correct to append 0xfff to
 *      make sure accesses to this first page raise a GP.  
 *
 * Results:
 *      The limit in bytes of the descriptor.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static INLINE uint32
Desc_ExpandedLimit(const Descriptor *d)
{
   return Desc_Gran(d) ? (Desc_GetLimit(d) << 12) | 0xfff : Desc_GetLimit(d);
}

/*
 *----------------------------------------------------------------------
 *
 * Desc_PackLimit --
 *
 *      Convert the limit of a descriptor into a 21-bit packed representation
 *
 * Results:
 *      packed representation of the limit
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static INLINE uint32
Desc_PackLimit(VA32 limit)
{
   if (limit < (1 << 20)) {
      return limit;
   } else {
      ASSERT((limit & 0xfff) == 0xfff);
      return (limit >> 12) | (1 << 20);
   }
}

/*
 *----------------------------------------------------------------------
 *
 * Desc_UnpackLimit --
 *
 *      Decode the representation of the limit as encoded by Desk_PackLimit
 *
 * Results:
 *      unpacked representation of the limit
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static INLINE VA
Desc_UnpackLimit(uint32 limit)
{
   ASSERT(limit < (1 << 21));
   if (limit & (1 << 20)) {
      limit = (limit << 12) | 0xfff;
   }
   return limit;
}

/*
 * The intel manual is misleading on expand down limit checks.  Valid offsets
 * actually range from the limit+1 to 0xffff 0xffffffff depending on the D
 * bit in the descriptor.  
 */
static INLINE Bool
Desc_InBoundsExpandDown(VA vaddr, VA limit, uint32 size, VA supremum)
{
   ASSERT(supremum == 0xffff || supremum == 0xffffffff);
   return vaddr > limit && vaddr <= supremum && size - 1 <= supremum - vaddr;
}


static INLINE Bool
Desc_InBoundsExpandUp(VA vaddr, VA limit, uint32 size)
{
   return vaddr <= limit && size - 1 <= limit - vaddr;
}


/*
 *----------------------------------------------------------------------
 *
 * Desc_InBounds and Desc_LegalAccess --
 *
 *      Check whether the access is within the bounds of the segment.
 *
 * Results:
 *      TRUE is the access is within the bounds of the segment, 
 *      FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
static INLINE Bool
Desc_InBounds(VA vaddr, VA limit, uint32 size, Bool expandUp, VA supremum)
{
   ASSERT(supremum == 0xffff || supremum == 0xffffffff);
   return expandUp ? Desc_InBoundsExpandUp(vaddr, limit, size)
                   : Desc_InBoundsExpandDown(vaddr, limit, size, supremum);
}


static INLINE Bool
Desc_LegalAccess(const Descriptor *d, VA vaddr, uint32 size)
{
   return Desc_InBounds(vaddr, Desc_ExpandedLimit(d), size, 
                        !DT_EXPAND_DOWN(*d), Desc_DB(d) ? 0xffffffff : 0xffff);
}


/*
 * Call Gates.
 */

typedef struct Gate {
   unsigned   offset_lo : 16;
   unsigned   segment   : 16;
   unsigned   params    : 5;
   unsigned   unused    : 3;
   unsigned   type      : 5;
   unsigned   DPL       : 2;
   unsigned   present   : 1;
   unsigned   offset_hi : 16;
} Gate;


#define GATE_OFFSET(_gate)       (((_gate).offset_hi << 16) | (_gate).offset_lo)

#define GATE_OFFSET_LO(_dw)      (((int)(_dw)) & 0xffff)
#define GATE_OFFSET_HI(_dw)      ((((int)(_dw)) >> 16) & 0xffff)

#define CALL_GATE                0x04
#define TASK_GATE                0x05
#define INTER_GATE               0x0e
#define TRAP_GATE                0x0f

#define GT_CALL(_gate)           (((_gate).type & 0x17) == 0x04)
#define GT_TASK(_gate)           (((_gate).type & 0x1f) == 0x05)
#define GT_INTR(_gate)           (((_gate).type & 0x17) == 0x06)
#define GT_TRAP(_gate)           (((_gate).type & 0x17) == 0x07)
#define GT_32BIT                 0x08
#define GT_32BIT_INTR            0xe
#define GT_32BIT_TRAP            0xf

/*
 * Descriptor Table Registers.
 */

#ifdef __GNUC__

/* 
 * Need to pack the DTR struct so the offset starts right after the
 * limit.  
 */
typedef struct DTR {
   uint16 limit;
   uint32 offset __attribute__ ((packed));
} DTR;

typedef struct DTR64 {
   uint16 limit;
   uint64 offset __attribute__ ((packed));
} DTR64;

#elif _MSC_VER

#pragma pack(push)
#pragma pack(2)

typedef struct {
   uint16 limit;
   uint32 offset;
} DTR;

typedef struct {
   uint16 limit;
   uint64 offset;
} DTR64;

#pragma pack(pop)

#else
# error
#endif

typedef union {
   DTR dtr;
   uint32 word[2];
} DTRWords;

typedef union {
   DTR64 dtr;
   uint64 word[2];
} DTRWords64;

#endif //_X86DESC_H_
