/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/
#define FILECODEINT 330

#include "driver-config.h"

#define EXPORT_SYMTAB

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/sched.h>
#ifdef KERNEL_2_2
#   include <linux/slab.h>
#else
#   include <linux/malloc.h>
#endif
#include <linux/poll.h>

#include <linux/smp.h>
#include <linux/smp_lock.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/mm.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/sockios.h>
#include "compat_sock.h"

#define __KERNEL_SYSCALLS__
#include <asm/io.h>

#include <linux/proc_fs.h>
#include <linux/file.h>
#ifdef __x86_64__
#include <asm/ioctl32.h>
#endif

#include "vm_oui.h"
#include "vnet.h"
#include "vnetInt.h"

#include "compat_uaccess.h"
#include "compat_kdev_t.h"
#include "compat_sched.h"
#include "compat_semaphore.h"
#include "vmnetInt.h"


/*
 * Initialization and creation routines from other files.
 * Putting them here reduces the need for so many header files.
 */

extern int VNetUserIf_Create(VNetPort **ret);
extern int VNetNetIf_Create(char *devName, VNetPort **ret, int hubNum);
extern int VNetBridge_Create(char *devName, VNetPort **ret);


/*
 *  Structure for cycle detection of host interfaces.
 */

typedef struct VNetInterface VNetInterface;

struct VNetInterface {
   char           name[8];
   int            myGeneration;
   VNetInterface *next;
};


/* this will let all multicast packets go through. */
uint8 AllMultiFilter[8] = { 0xff, 0xff, 0xff, 0xff,
                            0xff, 0xff, 0xff, 0xff };

/*
 * All jack->peer accesses are guarded by this lock.
 *
 * This lock is acquired for read from interrupt context: 
 * use write_lock_irqsave() to gain write access.
 *
 * If you are acquiring this lock for write, and you do
 * not have vnetStructureSemaphore already acquired,
 * it is most certainly bug.
 */
static rwlock_t vnetPeerLock = RW_LOCK_UNLOCKED;

/*
 * All concurrent changes to the network structure are 
 * guarded by this semaphore.
 *
 * For change to peer field you must own both
 * vnetStructureSemaphore and vnetPeerLock for write.
 */
DECLARE_MUTEX(vnetStructureSemaphore);

/*
 *  Device driver interface.
 */

static int  VNetFileOpOpen(struct inode *inode, struct file *filp);
static int  VNetFileOpClose(struct inode *inode, struct file *filp);
static unsigned int VNetFileOpPoll(struct file *filp, poll_table *wait);
static ssize_t  VNetFileOpRead(struct file *filp, char *buf, size_t count,
			       loff_t *ppos);
static ssize_t  VNetFileOpWrite(struct file *filp, const char *buf, size_t count,
			        loff_t *ppos);
static int  VNetFileOpIoctl(struct inode *inode, struct file *filp,
                            unsigned int iocmd, unsigned long ioarg);

static struct file_operations vnetFileOps;


/*
 * Utility functions
 */

static Bool VNetMulticastFilter(uint8 *destAddr, uint8 *ladrf);

static int  VNetPrintPIDs(VNetPort *port, char *buf);
static Bool VNetProcessOwnsPort(struct task_struct *p, VNetPort *port);


/*
 *----------------------------------------------------------------------
 *
 * VNetRegister --
 *
 *      (debugging support) Should be the first function of this file
 *
 * Results:
 *      
 *      Registers the module. 
 *      /sbin/ksyms -a | grep VNetRegister will return the base
 *      address of that function as loaded in the kernel. 
 * 
 *      Since this is the first function of the kernel module,
 *      every other symbol can be computing by adding the base
 *      to the output of nm.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int VNetRegister(int);

EXPORT_SYMBOL(VNetRegister);
EXPORT_SYMBOL(VNetHub_Alloc);
EXPORT_SYMBOL(VNetConnect);
EXPORT_SYMBOL(VNetDisconnect);
EXPORT_SYMBOL(VNetSend);
EXPORT_SYMBOL(VNetProc_MakeEntry);
EXPORT_SYMBOL(VNetProc_RemoveEntry);
EXPORT_SYMBOL(VNetPrintJack);

#ifdef MODULE_X86_64
static int
LinuxDriver_Ioctl32_Handler(unsigned int fd, unsigned int iocmd,
                            unsigned long ioarg, struct file * filp)
{
   int ret;
   lock_kernel();
   ret = -ENOTTY;
   if (filp && filp->f_op && filp->f_op->ioctl == VNetFileOpIoctl) {
      ret = VNetFileOpIoctl(filp->f_dentry->d_inode, filp, iocmd, ioarg);
   }
   unlock_kernel();
   return ret;
}

static int
register_ioctl32_handlers(void)
{
   int i;
   int retval;


   for (i = VNET_FIRST_CMD; i <= VNET_LAST_CMD; i++) {
      retval = register_ioctl32_conversion(i, LinuxDriver_Ioctl32_Handler);
      if (retval) {
         LOG(0, (KERN_WARNING "Fail to register ioctl32 conversion for cmd %d\n", i));
         return retval;
      }
   }
   return 0;
}

static void
unregister_ioctl32_handlers(void)
{  
   int i;
   int retval;

   for (i = VNET_FIRST_CMD; i <= VNET_LAST_CMD; i++) {
      retval = unregister_ioctl32_conversion(i);
      if (retval) {
         LOG(0, (KERN_WARNING "Fail to unregister ioctl32 conversion for cmd %d\n", i));
      }
   }
}
#else
static int
register_ioctl32_handlers(void)
{
   return 0;
}

static void
unregister_ioctl32_handlers(void)
{
}
#endif

int 
VNetRegister(int value)
{
   LOG(0, (KERN_WARNING "/dev/vmnet: VNetRegister called\n"));
   return 0;
}

/*
 * List of known ports. Use vnetStructureSemaphore for locking.
 */
static VNetPort *vnetAllPorts     = NULL;

/*
 *----------------------------------------------------------------------
 *
 * VNetAddPortToList --
 *
 *      Add port to list of known ports.
 *	Caller must own vnetStructureSemaphore.
 *
 * Results:
 *      
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static INLINE void
VNetAddPortToList(VNetPort *port)
{
   port->next = vnetAllPorts;
   vnetAllPorts = port;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetRemovePortFromList --
 *
 *      Remove port from list of known ports.
 *	Caller must own vnetStructureSemaphore.
 *
 * Results:
 *      
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static INLINE void
VNetRemovePortFromList(VNetPort *port)
{
   VNetPort **p;
   
   for (p = &vnetAllPorts; *p; p = &(*p)->next) {
      if (*p == port) {
         *p = port->next;
         break;
      }
   }
}


/*
 *----------------------------------------------------------------------
 *
 * init_module --
 *
 *      linux module entry point. Called by /sbin/insmod command
 * 
 * Results: 
 *      registers a device driver for a VNET major #.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int 
init_module(void)
{
   int retval;

   /*
    * First initialize everything, and as a last step register
    * vmnet device: immediately after registration anybody can
    * ask driver for anything.
    */   
   
   retval = VNetProc_Init();
   if (retval) {
      LOG(0, (KERN_NOTICE "/dev/vmnet: could not register proc fs\n"));
      return -ENOENT;
   }
   
   /* Initialize the file_operations structure. Because this code is always
      compiled as a module, this is fine to do it here and not in a static
      initializer --hpreg */
   memset(&vnetFileOps, 0, sizeof vnetFileOps);
   compat_fop_set_owner(&vnetFileOps);
   vnetFileOps.read = VNetFileOpRead;
   vnetFileOps.write = VNetFileOpWrite;
   vnetFileOps.poll = VNetFileOpPoll;
   vnetFileOps.ioctl = VNetFileOpIoctl;
   vnetFileOps.open = VNetFileOpOpen;
   vnetFileOps.release = VNetFileOpClose;

   retval = register_chrdev(VNET_MAJOR_NUMBER, "vmnet", &vnetFileOps);
   if (retval) {
      LOG(0, (KERN_NOTICE "/dev/vmnet: could not register major device %d\n",
	      VNET_MAJOR_NUMBER));
      VNetProc_Cleanup();
      return -ENOENT;
   }

   retval = register_ioctl32_handlers();
   if (retval) {
      return retval;
   }
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * cleanup_module --
 *
 *      Called by /sbin/rmmod
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

void
cleanup_module(void)
{
   int retval;
   unregister_ioctl32_handlers();
   retval =  unregister_chrdev(VNET_MAJOR_NUMBER, "vmnet");
   if (retval != 0 ) {
      LOG(0, (KERN_WARNING "/dev/vmnet: could not unregister major device %d\n",
	      VNET_MAJOR_NUMBER));
   }
   VNetProc_Cleanup();
}


/*
 *----------------------------------------------------------------------
 *
 * VNetIncrModCount --
 *
 *      Increment or decrement the module use count.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

void
VNetIncrModCount(int delta)
{
   if (delta > 0) {
      compat_mod_inc_refcount;
   } else if (delta < 0) {
      compat_mod_dec_refcount;
   }
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpOpen --
 *
 *      The virtual network's open file operation. Open a connection to
 *      this virtual network. (Like plugging a cable into this virtual
 *      hub.)
 *
 * Results: 
 *      Errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetFileOpOpen(struct inode  *inode,
               struct file   *filp)
{
   VNetPort *port;
   VNetJack *hubJack;
   int hubNum;
   int retval;
   
   LOG(1, (KERN_DEBUG "/dev/vmnet: open called by PID %d (%s)\n",
           current->pid, current->comm));

   /*
    * Allocate port
    */
   
   retval = VNetUserIf_Create(&port);
   if (retval) {
      return -retval;
   }
   
   /*
    * Connect to hub.
    */
   
   hubNum = minor(inode->i_rdev);
   if ((hubNum < 0) || (hubNum >= VNET_NUM_HUBS)) {
      VNetFree(&port->jack);
      return -ENODEV;
   }

   hubJack = VNetHub_Alloc(hubNum);
   if (!hubJack) {
      VNetFree(&port->jack);
      return -EBUSY;
   }

   down(&vnetStructureSemaphore);
   retval = VNetConnect(&port->jack, hubJack);
   if (retval) {
      up(&vnetStructureSemaphore);
      VNetFree(&port->jack);
      VNetFree(hubJack);
      return retval;
   }
   
   VNetAddPortToList(port);
   up(&vnetStructureSemaphore);
   
   /*
    * Store away jack in file pointer private field for later use.
    */
   
   filp->private_data = port;
   
   LOG(1, (KERN_DEBUG "/dev/vmnet: port on hub %d successfully opened\n", hubNum));
   
   compat_mod_inc_refcount;

   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpClose --
 *
 *      The virtual network's close file operation. Unplugging the cable.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int
VNetFileOpClose(struct inode  *inode,
                struct file   *filp)
{
   VNetPort *port = (VNetPort*)filp->private_data;
   VNetJack *peer;
   
   if (!port) {
      LOG(1, (KERN_DEBUG "/dev/vmnet: bad file pointer on close\n"));
      return -EBADF;
   }

   down(&vnetStructureSemaphore);
   peer = VNetDisconnect(&port->jack);
   VNetRemovePortFromList(port);
   up(&vnetStructureSemaphore);

   VNetFree(&port->jack);
   VNetFree(peer);

   compat_mod_dec_refcount;

   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpRead --
 *
 *      The virtual network's read file operation.
 *
 * Results: 
 *      On success the len of the packet received,
 *      else if no packet waiting and nonblocking 0,
 *      else -errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

ssize_t
VNetFileOpRead(struct file  *filp,
               char         *buf,
               size_t        count,
	       loff_t	    *ppos)
{
   VNetPort *port = (VNetPort*)filp->private_data;
   
   if (!port) {
      LOG(1, (KERN_DEBUG "/dev/vmnet: bad file pointer on read\n"));
      return -EBADF;
   }
   
   if (!port->fileOpRead) {
      return -EPERM;
   }
   
   return port->fileOpRead(port, filp, buf, count);
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpWrite --
 *
 *      The virtual network's write file operation.
 *
 * Results: 
 *      On success the count of bytes written else errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

ssize_t
VNetFileOpWrite(struct file   *filp,
                const char    *buf,
                size_t         count,
		loff_t	      *ppos)
{
   VNetPort *port = (VNetPort*)filp->private_data;

   if (!port) {
      LOG(1, (KERN_DEBUG "/dev/vmnet: bad file pointer on write\n"));
      return -EBADF;
   }
   
   if (!port->fileOpWrite) {
      return -EPERM;
   }
   
   return port->fileOpWrite(port, filp, buf, count);
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpPoll --
 *
 *      The virtual network's file select operation.
 *
 * Results: 
 *      Return 1 if success, else sleep and return 0.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static unsigned int
VNetFileOpPoll(struct file *filp,
	       poll_table  *wait)
{
   VNetPort *port = (VNetPort*)filp->private_data;

   if (!port) {
      LOG(1, (KERN_DEBUG "/dev/vmnet: bad file pointer on poll\n"));
      return -EBADF;
   }
   
   if (!port->fileOpPoll) {
      return -EPERM;
   }
   
   return port->fileOpPoll(port, filp, wait);
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpIoctl --
 *
 *      The virtual network's ioctl file operation. This is used for
 *      setup of the connection. Currently supported commands are
 *      (taken from sockios.h):
 *
 *      SIOCGIFADDR - get ethernet address          - ioarg OUT: 6 bytes
 *      SIOCSIFADDR - set ethernet address          - ioarg IN:  6 bytes
#ifdef VNET_API_DEPRECATED
 *      SIOCGIFFLAGS - get flags                    - ioarg OUT  4 bytes
#endif
 *      SIOCSIFFLAGS - set flags                    - ioarg IN:  4 bytes
 *      SIOCGIFBR - get bridging status             - ioarg OUT: 4 bytes
 *      SIOCSIFBR - set bridge peer interface       - ioarg IN:  8 bytes
 *
 *      Private ioctl calls, taken from device-private ioctl space 
 *      in sockios.h, and defined in includes/vm_oui.h:
 *
 *      SIOCSKEEP (0x89F0) - set dealloc flag       - ioarg IN:  1 byte
 *      SIOCGKEEP (0x89F1) - get dealloc flag       - ioarg OUT: 1 byte
 *         By default all hubs (and their associated interfaces) are 
 *         deallocated from memory when their last open port is closed.  
 *         Setting this true "keeps" the hub and thus we will not 
 *         deallocate it until the module is unloaded.
 *
 *      SIOCSLADRF (0x89F2) - set logical address filter (for 
 *         filtering multicast packets)             - ioarg IN:  8 bytes
 *
 *
 *      Supported flags are (taken from if.h):
 *
 *      IFF_UP - ready to receive packets             - OFF by default
 *      IFF_BROADCAST - receive broadcast packets     - OFF by default
 *      IFF_DEBUG - turn on debugging                 - OFF by default
 *      IFF_PROMISC - promiscuous mode                - OFF by default
 *      IFF_MULTICAST - receive multicast packets     - OFF by default
 *      IFF_ALLMULTI - receive all multicast packets 
 *            (like IFF_PROMISC but with multicast)   - OFF by default
 *
 * Results: 
 *      On success 0 else errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetFileOpIoctl(struct inode   *inode,
                struct file    *filp,
                unsigned int    iocmd,
                unsigned long   ioarg)
{
   VNetPort *port = (VNetPort*)filp->private_data;
   VNetJack *peer;
   VNetPort *new;
   char name[8];
   int retval, retval2;
   VNet_SetMacAddrIOCTL macAddr;
   
   if (!port) {
      LOG(1, (KERN_DEBUG "/dev/vmnet: bad file pointer on ioctl\n"));
      return -EBADF;
   }
   
   // sprintf(vnetHub[hubNum]->devName, "vmnet%d", hubNum);

   switch (iocmd) {
   case SIOCSIFBR:
   case SIOCPORT:   
   case SIOCBRIDGE:
   case SIOCNETIF:
      if (iocmd == SIOCPORT) {
         retval = VNetUserIf_Create(&new);
      } else {
         if (copy_from_user(name, (void *)ioarg, sizeof name)) {
            return -EFAULT;
         }

         name[sizeof name - 1] = 0;

         if (iocmd == SIOCNETIF) {
            retval = VNetNetIf_Create(name, &new, minor(inode->i_rdev));
         } else {
            if (!capable(CAP_NET_RAW)) {
               return -EACCES;
            }
   
            retval = VNetBridge_Create(name, &new);
         }
      }
      if (retval) {
         return retval;
      }

      /*
       *  OK this is tricky. Try and connect the new port to the old
       *  ports peer. However we need to save enough info so that we can
       *  reconnect with the old port if a cycle is detected.
       */
      
      down(&vnetStructureSemaphore);
      peer = VNetDisconnect(&port->jack);

      retval = VNetConnect(&new->jack, peer);
      if (retval) {
         retval2 = VNetConnect(&port->jack, peer);
         up(&vnetStructureSemaphore);
         VNetFree(&new->jack);
         if (retval2) {
            // assert xxx redo this
            LOG(1, (KERN_NOTICE "/dev/vmnet: cycle on connect failure\n"));
            return -EBADF;
         }
         return retval;
      }
      VNetAddPortToList(new);
      filp->private_data = new;
      VNetRemovePortFromList(port);

      up(&vnetStructureSemaphore);

      VNetFree(&port->jack);
      break;

   case SIOCGIFBR:
      {
         int flags;

	 read_lock(&vnetPeerLock);
         flags = VNetIsBridged(&port->jack);
	 read_unlock(&vnetPeerLock);

         retval = compat_put_user(flags, (int *)ioarg);
	 if (retval < 0) {
            return retval;
	 }
      }
      break;
      
   case SIOCGIFADDR:
      if (copy_to_user((void *)ioarg, port->paddr, ETH_ALEN)) {
         return -EFAULT;
      }
      break;
      
   case SIOCSIFADDR:
      return -EFAULT;
      
   case SIOCSLADRF:
      if (copy_from_user(port->ladrf, (void *)ioarg, 8)) {
         return -EFAULT;
      }
      break;

#ifdef VNET_API_DEPRECATED
   /* We can't think of a good reason to use this ioctl. --gustav,hpreg */
   case SIOCGIFFLAGS:
      retval = compat_put_user(port->flags, (uint32 *)ioarg);
      if (retval < 0) {
         return retval;
      }
      break;
      
#endif
   case SIOCSIFFLAGS:
      if (copy_from_user(&port->flags, (void *)ioarg, sizeof port->flags)) {
         return -EFAULT;
      }
      port->flags = ((port->flags
                      & (IFF_UP|IFF_BROADCAST|IFF_DEBUG
                         |IFF_PROMISC|IFF_MULTICAST|IFF_ALLMULTI))
                     | IFF_RUNNING);
      /* XXX process queued packets? */
      break;

   case SIOCSETMACADDR:
      if (copy_from_user(&macAddr, (void *)ioarg, sizeof macAddr)) {
         return -EFAULT;
      }
      
      switch (macAddr.version) {
      case 1:
         if (macAddr.flags & VNET_SETMACADDRF_UNIQUE) {
	    if (VMX86_IS_VIRT_ADAPTER_MAC(macAddr.addr)) {
	       return -EBUSY;
	    }
            return VNetSetMACUnique(port, macAddr.addr);
         }
         memcpy(port->paddr, macAddr.addr, ETH_ALEN);
         break;
      default:
         return -EINVAL;
         break;
      }
      break;
   case SIOCGETAPIVERSION: 
      retval = compat_put_user(VNET_API_VERSION, (int *)ioarg);
      if (retval < 0) {
	 return retval;
      }
      break;
   default:
      if (!port->fileOpIoctl) {
         return -ENOIOCTLCMD;
      }
      return port->fileOpIoctl(port, filp, iocmd, ioarg);
      break;
   }
   
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetPacketMatch --
 *
 *      Determines whether the packet should be given to the interface.
 *
 * Results: 
 *      TRUE if the pasket is OK for this interface, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
VNetPacketMatch(uint8   *destAddr,
                uint8   *ifAddr,
		uint8   *ladrf,
                uint32   flags)
{
   static uint8 broadcast[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
   return ((flags & IFF_PROMISC)
           || MAC_EQ(destAddr, ifAddr)
           || ((flags & IFF_BROADCAST) && MAC_EQ(destAddr, broadcast))
           || ((destAddr[0] & 0x1) && 
	       (flags & IFF_ALLMULTI || // allow all multi
		     (flags & IFF_MULTICAST && 
		      VNetMulticastFilter(destAddr, ladrf)))));
}

/*
 *----------------------------------------------------------------------
 *
 * VNetMulticastFilter --
 *
 *      Utility function that filters multicast packets according
 *      to a 64-bit logical address filter (like the one on the
 *      lance chipset).  AllMultiFilter lets all packets through.
 *
 *      We generate a hash value from the destination MAC address
 *      and see if it's in our filter.  Broadcast packets have
 *      already OK'd by PacketMatch, so we don't have to worry
 *      about that.
 *
 *      (This is in the green AMD "Ethernet Controllers" book, 
 *      page 1-53.)
 *
 * Results:
 *      TRUE if packet is in filter, FALSE if not.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

#define CRC_POLYNOMIAL_BE 0x04c11db7UL  /* Ethernet CRC, big endian */

Bool
VNetMulticastFilter(uint8 *destAddr,
		    uint8 *ladrf)
{
   uint16 hashcode;
   int32 crc, poly = CRC_POLYNOMIAL_BE;
   int j, bit, byte;
   
   crc = 0xffffffff;                  /* init CRC for each address */
   for (byte=0; byte<6; byte++) {     /* for each address byte */
      /* process each address bit */ 
      for (bit = *destAddr++,j=0;j<8;j++, bit>>=1) {
	 crc = (crc << 1) ^ ((((crc<0?1:0) ^ bit) & 0x01) ? poly : 0);
      }
   }
   hashcode = (crc & 1);              /* hashcode is 6 LSb of CRC ... */
   for (j=0; j<5; j++) {                /* ... in reverse order. */
      hashcode = (hashcode << 1) | ((crc>>=1) & 1);
   }                                      
   
   byte = hashcode >> 3;              /* bit[3-5] -> byte in filter */
   bit = 1 << (hashcode & 0x07);      /* bit[0-2] -> bit in byte */
   if (ladrf[byte] & bit) {
      return TRUE;
   } else {
      return FALSE;
   }
}


/*
 *----------------------------------------------------------------------
 *
 * VNet_MakeMACAddress --
 *
 *      Generate a unique MAC address and assign it to the given port.
 *      The address will be in the range:
 *
 *      VMX86_STATIC_OUI:e0:00:00 - VMX86_STATIC_OUI:ff:ff:ff
 *
 * Results: 
 *      errno.
 *
 * Side effects:
 *      The address is changed.
 *
 *----------------------------------------------------------------------
 */

int
VNet_MakeMACAddress(VNetPort *port)
{
   uint8 paddr[ETH_ALEN];
   int conflict;
   int maxTries = 1000;

   memset(paddr, 0, ETH_ALEN);

   do {
      VMX86_GENERATE_RANDOM_MAC(paddr);

      conflict = VNetSetMACUnique(port, paddr);
      
      /*
       * We don't have to check for conflicts with the virtual
       * host adapters since they are in the range 
       * c0:00:00-c0:00:FF.
       */
   } while (maxTries-- > 0 && conflict);
   
   return conflict;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetConnect --
 *
 *      Connect 2 jacks.
 *	vnetStructureSemaphore must be held.
 *
 * Results: 
 *      errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetConnect(VNetJack *jack1,
            VNetJack *jack2)
{
   static int vnetGeneration = 0;
   Bool foundCycle;
   unsigned long flags;
   
   vnetGeneration++;

   foundCycle = VNetCycleDetect(jack1, vnetGeneration);
   if (foundCycle) {
      return -EDEADLK;
   }
   
   foundCycle = VNetCycleDetect(jack2, vnetGeneration);
   if (foundCycle) {
      return -EDEADLK;
   }
   
   /*
    * Synchronize with peer readers
    */
   write_lock_irqsave(&vnetPeerLock, flags);
   jack1->peer = jack2;
   jack2->peer = jack1;
   write_unlock_irqrestore(&vnetPeerLock, flags);

   if (jack2->numPorts) {
      VNetPortsChanged(jack1);
   }

   if (jack1->numPorts) {
      VNetPortsChanged(jack2);
   }

   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetDisconnect --
 *
 *	Disconnect 2 jacks.
 *	vnetStructureSemaphore must be held.
 *
 * Results: 
 *      Return the peer jack.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

VNetJack *
VNetDisconnect(VNetJack *jack)
{
   VNetJack *peer;
   unsigned long flags;
   
   write_lock_irqsave(&vnetPeerLock, flags);
   peer = jack->peer;
   if (!peer) {
      write_unlock_irqrestore(&vnetPeerLock, flags);
      return NULL;
   }
   jack->peer = NULL;
   peer->peer = NULL;
   write_unlock_irqrestore(&vnetPeerLock, flags);

   if (peer->numPorts) {
      VNetPortsChanged(jack);
   }

   if (jack->numPorts) {
      VNetPortsChanged(peer);
   }

   return peer;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetCycleDetectIf --
 *
 *      Perform the cycle detect alogorithm for this generation on a
 *      specific interface. This could be a bridged interface, host
 *      interface or both.
 *
 * Results: 
 *      TRUE if a cycle was detected, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
VNetCycleDetectIf(char *name,
                  int   generation)
{
   static VNetInterface *vnetInterfaces = 0;
   VNetInterface *p;

   for (p=vnetInterfaces; p; p=p->next) {
      if (!strcmp(name, p->name)) {
         if (p->myGeneration == generation) {
            return TRUE;
         } else {
            p->myGeneration = generation;
            return FALSE;
         }
      }
   }
   
   p = kmalloc(sizeof *p, GFP_USER);
   if (!p) {
      // assert
      return TRUE;
   }
   
   memcpy(p->name, name, sizeof p->name);
   p->myGeneration = generation;
   p->next = vnetInterfaces;
   vnetInterfaces = p;

   return FALSE;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetSend --
 *
 *      Send a packet through this jack. Note, the packet goes to the
 *      jacks peer.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The skb is no longer owned by us.
 *
 *----------------------------------------------------------------------
 */

void
VNetSend(VNetJack        *jack,
         struct sk_buff  *skb)
{
   read_lock(&vnetPeerLock);
   if (jack && jack->peer && jack->peer->rcv) {
      jack->peer->rcv(jack->peer, skb);
   } else {
      dev_kfree_skb(skb);
   }
   read_unlock(&vnetPeerLock);
}


/*
 *----------------------------------------------------------------------
 *
 * VNetSetMACUnique --
 *
 *      Verify that MAC address is not used by other ports.
 *
 * Results:
 *      
 *      0      if address is unique. Port's paddr is updated.
 *      -EBUSY if address is already in use. Port's paddr is unchanged.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetSetMACUnique(VNetPort *port, 
		 const uint8 mac[ETH_ALEN]) {
   VNetPort *p;
   
   down(&vnetStructureSemaphore);
   for (p = vnetAllPorts; p; p = p->next) {
      if ((p != port) && MAC_EQ(p->paddr, mac)) {
         up(&vnetStructureSemaphore);
         return -EBUSY;
      }
   }
   memcpy(port->paddr, mac, ETH_ALEN);
   up(&vnetStructureSemaphore);
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetPrintJack --
 *
 *      Print info about the jack to a buffer.
 *
 * Results:
 *      Length of the write.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetPrintJack(VNetJack  *jack,
              char      *buf)
{
   int len = 0;
   
   read_lock(&vnetPeerLock);
   if (!jack->peer) {
      len += sprintf(buf+len, "connected not ");
   } else {
      len += sprintf(buf+len, "connected %s ", jack->peer->name);
   }
   read_unlock(&vnetPeerLock);

   return len;
}
              

/*
 *----------------------------------------------------------------------
 *
 * VNetPrintPort --
 *
 *      Print info about the port to a buffer.
 *
 * Results:
 *      Length of the write.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetPrintPort(VNetPort  *port,
              char      *buf)
{
   int len = 0;

   len += VNetPrintJack(&port->jack, buf+len);

   len += VNetPrintPIDs(port, buf+len);

   len += sprintf(buf+len, "mac %02x:%02x:%02x:%02x:%02x:%02x ",
                  port->paddr[0], port->paddr[1], port->paddr[2], 
                  port->paddr[3], port->paddr[4], port->paddr[5]); 

   len += sprintf(buf+len, "ladrf %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x ",
                  port->ladrf[0], port->ladrf[1], port->ladrf[2],
                  port->ladrf[3], port->ladrf[4], port->ladrf[5],
                  port->ladrf[6], port->ladrf[7]);

   len += sprintf(buf+len, "flags IFF_RUNNING");
   
   if (port->flags & IFF_UP) {
      len += sprintf(buf+len, ",IFF_UP");      
   }

   if (port->flags & IFF_BROADCAST) {
      len += sprintf(buf+len, ",IFF_BROADCAST");      
   }

   if (port->flags & IFF_DEBUG) {
      len += sprintf(buf+len, ",IFF_DEBUG");      
   }

   if (port->flags & IFF_PROMISC) {
      len += sprintf(buf+len, ",IFF_PROMISC");      
   }

   if (port->flags & IFF_MULTICAST) {
      len += sprintf(buf+len, ",IFF_MULTICAST");      
   }

   if (port->flags & IFF_ALLMULTI) {
      len += sprintf(buf+len, ",IFF_ALLMULTI");      
   }

   len += sprintf(buf+len, " ");
   
   return len;
}
              

/*
 *----------------------------------------------------------------------
 *
 * VNetPrintPIDs --
 *
 *      Return a list of PID that have this port open.
 *
 * Results: 
 *      The number of characters of the list, and a string with a
 *      comma separated list of PIDs.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetPrintPIDs(VNetPort *port,  // IN: the file in question
              char     *buf)   // OUT: the buffer to fill in
{
   struct task_struct *p;
   int len = 0;
   int seen = 0;
   
   len += sprintf(buf+len, "pids ");

   /*
    * We do not handle properly threads which do use CLONE_PID, but
    * do not use CLONE_FILES
    */

   read_lock(&tasklist_lock);
   for_each_process(p) {
      if (VNetProcessOwnsPort(p, port)) {
         len += sprintf(buf+len, "%d", p->pid);
         if (seen) {
            len += sprintf(buf+len, ",");
         } else {
            seen = 1;
         }
      }
   }
   read_unlock(&tasklist_lock);

   if (!seen) {
      len += sprintf(buf+len, "{}");
   }

   len += sprintf(buf+len, " ");
   
   return len;
}

/*
 *----------------------------------------------------------------------
 *
 * VNetProcessOwnsPort --
 *
 *      Return whether a process has the port open.
 *
 * Results: 
 *      TRUE if the process has the file open, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
VNetProcessOwnsPort(struct task_struct  *p,
                    VNetPort            *port)
{
   struct file *file;
   int fd;
   int maj;
   
   if (!p->files) {
      return FALSE;
   }
   
   for (fd = 0; fd < p->files->max_fds; fd++) {
#ifdef KERNEL_2_3_99
      file = fcheck_files(p->files, fd);
#else
      file = fcheck_task(p, fd);
#endif
      if (file && file->f_dentry) {
         maj = major(file->f_dentry->d_inode->i_rdev);
         if ((maj == VNET_MAJOR_NUMBER) && (file->private_data == port)) {
            return TRUE;
         }
      }
   }

   return FALSE;
}
