/* **********************************************************
 * Copyright 2003 VMware, Inc.  All rights reserved.
 * -- VMware Confidential
 * **********************************************************/

/*
 * vm_basic_asm.h
 *
 *	Basic asm macros
 */

#ifndef _VM_BASIC_ASM_H_
#define _VM_BASIC_ASM_H_

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_VMM
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_DISTRIBUTE
#include "includeCheck.h"

#include "vm_basic_types.h"

#if defined(MODULE_X86_64) && defined(USERLEVEL) && defined(_MSC_VER)
#ifdef __cplusplus
extern "C" {
#endif
#include "ioaccess.h"
#ifdef __cplusplus
}
#endif
#endif

#ifdef MODULE_X86_64
#include "vm_basic_asm_x86_64.h"
#else
#include "vm_basic_asm_x86.h"
#endif


#ifdef __GNUC__ // {

/*
 * Checked against the Intel manual and GCC --hpreg
 *
 * volatile because reading from port can modify the state of the underlying
 * hardware.
 *
 * Note: The undocumented %z construct doesn't work (internal compiler error)
 *       with gcc-2.95.1
 */

#define __GCC_IN(s, type, name) \
static INLINE type              \
name(uint16 port)               \
{                               \
   type val;                    \
                                \
   __asm__ __volatile__(        \
      "in" #s " %w1, %0"        \
      : "=a" (val)              \
      : "Nd" (port)             \
   );                           \
                                \
   return val;                  \
}

__GCC_IN(b, uint8, INB)
__GCC_IN(w, uint16, INW)
__GCC_IN(l, uint32, IN32)


/*
 * Checked against the Intel manual and GCC --hpreg
 *
 * Note: The undocumented %z construct doesn't work (internal compiler error)
 *       with gcc-2.95.1
 */

#define __GCC_OUT(s, s2, port, val) do { \
   __asm__(                              \
      "out" #s " %" #s2 "1, %w0"         \
      :                                  \
      : "Nd" (port), "a" (val)           \
   );                                    \
} while (0)

#define OUTB(port, val) __GCC_OUT(b, b, port, val)
#define OUTW(port, val) __GCC_OUT(w, w, port, val)
#define OUT32(port, val) __GCC_OUT(l, , port, val)


#define GET_CURRENT_EIP(_eip) \
      __asm__ __volatile("call 0\n\tpopl %0" : "=r" (_eip): );


/*
 * Checked against the Intel manual and GCC --hpreg
 * 
 * Need __volatile__ and "memory" since CPUID has a synchronizing effect.
 * The CPUID may also change at runtime (APIC flag, etc).
 *
 */

static INLINE void
__GET_CPUID(int ax, uint32 *regs)
{
   __asm__ __volatile__(
      "cpuid"
      : "=a" (regs[0]), "=b" (regs[1]), "=c" (regs[2]), "=d" (regs[3])
      : "a" (ax)
      : "memory"
   );
}


#elif _MSC_VER // } {
#ifdef MODULE_X86_64
static INLINE  uint8
INB(uint16 port) 
{
   return __inbyte(port);
}
static INLINE void
OUTB(uint16 port, uint8 value)
{
   __outbyte(port, value);
}
static INLINE uint16
INW(uint16 port)
{
   return __inword(port);
}
static INLINE void
OUTW(uint16 port, uint16 value)
{
   __outword(port, value);
}
static INLINE  uint32
IN32(uint16 port) 
{
   return __indword(port);
}
static INLINE void 
OUT32(uint16 port, uint32 value)
{
   __outdword(port, value);
}
#else
static INLINE  uint8
INB(uint16 port) 
{ 
   uint8 val;
   __asm mov dx,port
   __asm in al,dx
   __asm mov val, al
   return val;
}
static INLINE void
OUTB(uint16 port, uint8 value)
{
  __asm mov dx, port
  __asm mov al, value
  __asm out dx,al
}

static INLINE  uint16
INW(uint16 port) 
{ 
   uint16 val;
   __asm mov dx,port
   __asm in ax,dx
   __asm mov val, ax
   return val;
}
static INLINE void 
OUTW(uint16 port, uint16 value)
{
  __asm mov dx, port
  __asm mov ax, value
  __asm out dx,ax
}

static INLINE  uint32
IN32(uint16 port) 
{ 
   uint32 val;
   __asm mov dx,port
   __asm in eax,dx
   __asm mov val, eax
   return val;
}
static INLINE void 
OUT32(uint16 port, uint32 value)
{
  __asm mov dx, port
  __asm mov eax, value
  __asm out dx,eax
}


#ifdef NEAR
#undef NEAR
#endif

#define GET_CURRENT_EIP(_eip) { \
   __asm call NEAR PTR $+5 \
   __asm pop eax \
   __asm mov _eip, eax \
}

static INLINE void
__GET_CPUID(int input, uint32 *regs)
{
   __asm push esi
   __asm push ebx
   __asm push ecx
   __asm push edx

   __asm mov  eax, input
   __asm mov  esi, regs
   __asm _emit 0x0f __asm _emit 0xa2
   __asm mov 0x0[esi], eax
   __asm mov 0x4[esi], ebx
   __asm mov 0x8[esi], ecx
   __asm mov 0xC[esi], edx

   __asm pop edx
   __asm pop ecx
   __asm pop ebx
   __asm pop esi

}

#endif /* MODULE_X86_64 */
#else // }
#error 
#endif


#define GET_CPUID(_ax,_bx,_cx,_dx) { \
   uint32 regs[4];                   \
   __GET_CPUID(_ax,regs);            \
   _ax = regs[0];                    \
   _bx = regs[1];                    \
   _cx = regs[2];                    \
   _dx = regs[3];                    \
}


#ifdef _MSC_VER   
#ifdef MODULE_X86_64
EXTERN int ffs(uint32 bitVector);
#else
static INLINE int
ffs(uint32 bitVector)
{
   int idx;
   if (!bitVector) {
      return 0;
   }
   __asm bsf eax, bitVector
   __asm mov idx, eax
   return idx+1;
}
#endif /* MODULE_X86_64 */
#endif

#ifdef __GNUC__
static INLINE void *
uint16set(void *dst, uint16 val, size_t count)
{
   int dummy0;
   int dummy1;
   __asm__ __volatile__("\t"
                        "cld"            "\n\t"
                        "rep ; stosw"    "\n"
                        : "=c" (dummy0), "=D" (dummy1)
                        : "0" (count), "1" (dst), "a" (val)
                        : "memory", "cc"
      );

   return dst;
}

static INLINE void *
uint32set(void *dst, uint32 val, size_t count)
{
   int dummy0;
   int dummy1;
   __asm__ __volatile__("\t"
                        "cld"            "\n\t"
                        "rep ; stosl"    "\n"
                        : "=c" (dummy0), "=D" (dummy1)
                        : "0" (count), "1" (dst), "a" (val)
                        : "memory", "cc"
      );

   return dst;
}

#elif _MSC_VER
#ifdef MODULE_X86_64
#else
static INLINE void *
uint16set(void *dst, uint16 val, size_t count)
{
   __asm { pushf;
           mov ax, val;
           mov ecx, count;
           mov edi, dst;
           cld;
           rep stosw;
           popf;
   }
   return dst;
}

static INLINE void *
uint32set(void *dst, uint32 val, size_t count)
{
   __asm { pushf;
           mov eax, val;
           mov ecx, count;
           mov edi, dst;
           cld;
           rep stosd;
           popf;
   }
   return dst;
}
#endif /* MODULE_X86_64 */
#else
#error "No compiler defined for uint*set"
#endif


/*
 *-----------------------------------------------------------------------------
 *
 * Bswap --
 *
 *    Swap the 4 bytes of "v" as follows: 3210 -> 0123.
 *
 *-----------------------------------------------------------------------------
 */

#ifdef __GNUC__ // {
static INLINE uint32
Bswap(uint32 v)
{
   /* Checked against the Intel manual and GCC --hpreg */
   __asm__(
      "bswap %0" 
      : "=r" (v)
      : "0" (v)
   );
   return v;
}
#endif // }

#ifdef __GNUC__ // {
/*
 * COMPILER_MEM_BARRIER prevents the compiler from re-ordering memory
 * references accross the barrier.  NOTE: It does not generate any
 * instruction, so the CPU is free to do whatever it wants to...
 */
#define COMPILER_MEM_BARRIER() __asm__ __volatile__ ("": : :"memory")
#endif // }

#endif
